import logging
from json import JSONEncoder, dumps

from flask import current_app, make_response, request

__all__ = (
    "serialize",
    "jsonify",
)

log = logging.getLogger(__name__)


def serialize(rv):
    log.debug("Serializing output")

    if rv is None or (isinstance(rv, str) and not len(rv)):
        log.info("No content")
        rv = make_response("", 204)
    elif (
        isinstance(rv, current_app.response_class)
        or callable(rv)
        or isinstance(rv, str)
    ):
        ...
    else:
        log.info("Serializing")
        rv = jsonify(rv)
        if request.method == "POST":
            make_response(rv, 201)

    return rv


def jsonify(*args, **kwargs):
    if args and kwargs:
        raise TypeError("jsonify() behavior undefined when passed both args and kwargs")
    elif len(args) == 1:  # single args are passed directly to dumps()
        data = args[0]
    else:
        data = args or kwargs

    pretty_print = bool(
        request.args.get(
            "pretty-print", current_app.config["JSONIFY_PRETTYPRINT_REGULAR"]
        )
    )

    indent = None
    separators = (",", ":")
    cls = current_app.json_encoder or JSONEncoder

    if pretty_print is True and request.is_xhr is False:
        indent = 2
        separators = (", ", ": ")

    if hasattr(request, "operation") and request.operation.produces:
        mime_type = request.operation.produces[0]
    elif "JSONIFY_MIMETYPE" in current_app.config:
        mime_type = current_app.config["JSONIFY_MIMETYPE"]
    else:
        mime_type = "application/json; charset=utf-8"

    json_str = dumps(data, indent=indent, separators=separators, cls=cls) + "\n"
    json_str.encode("utf-8")

    return current_app.response_class(json_str, mimetype=mime_type)



import datetime
import logging
import os
import subprocess
import threading
import time

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql import text

from Pegasus import braindump, user
from Pegasus.db import connection
from Pegasus.db.ensembles import (
    EMError,
    Ensembles,
    EnsembleStates,
    EnsembleWorkflowStates,
)
from Pegasus.db.schema import MasterWorkflow, MasterWorkflowstate
from Pegasus.service.ensembles import emapp

log = logging.getLogger(__name__)


def pathfind(exe):
    PATH = os.getenv("PATH", "/bin:/usr/bin:/usr/local/bin")
    PATH = PATH.split(":")
    for prefix in PATH:
        exepath = os.path.join(prefix, exe)
        if os.path.isfile(exepath):
            return exepath
    raise EMError("%s not found on PATH" % exe)


def get_bin(name, exe):
    # Try to find NAME/bin using 1) NAME env var, 2) NAME config
    # variable, 3) PATH env var
    exepath = None

    HOME = os.getenv(name, emapp.config.get(name, None))

    if HOME is not None:
        if not os.path.isdir(HOME):
            raise EMError("{} is not a directory: {}".format(name, HOME))
        BIN = os.path.join(HOME, "bin")
        if not os.path.isdir(BIN):
            raise EMError("{}/bin is not a directory: {}".format(name, BIN))
        exepath = os.path.join(BIN, exe)

    exepath = exepath or pathfind(exe)

    if not os.path.isfile(exepath):
        raise EMError("{} not found: {}".format(exe, exepath))

    BIN = os.path.dirname(exepath)

    return BIN


def get_pegasus_bin():
    return get_bin("PEGASUS_HOME", "pegasus-plan")


def get_condor_bin():
    return get_bin("CONDOR_HOME", "condor_submit_dag")


def check_environment():
    # Make sure Pegasus and Condor are on the PATH, or PEGASUS_HOME
    # and CONDOR_HOME are set in the environment or configuration
    get_pegasus_bin()
    get_condor_bin()


def get_script_env():
    PEGASUS_BIN = get_pegasus_bin()
    CONDOR_BIN = get_condor_bin()

    # Add pegasus bin dirs to PATH
    env = dict(os.environ)
    PATH = env.get("PATH", "/bin:/usr/bin:/usr/local/bin")
    PATH = PEGASUS_BIN + ":" + CONDOR_BIN + ":" + PATH
    env["PATH"] = PATH

    return env


def runscript(script, cwd=None, env=None):
    # Make sure the cwd is OK
    if cwd is not None and not os.path.isdir(cwd):
        raise EMError("Working directory does not exist: %s" % cwd)

    if env is None:
        env = dict(os.environ)

    p = subprocess.Popen(script, shell=True, env=env, cwd=cwd)

    rc = p.wait()

    if rc != 0:
        raise EMError("Script failed with exitcode %d" % rc)


def forkscript(script, pidfile=None, cwd=None, env=None):
    # This does a double fork to detach the process from the python
    # interpreter so that we don't have to call wait() on it

    # Make sure the cwd is OK
    if cwd is not None and not os.path.isdir(cwd):
        raise EMError("Working directory does not exist: %s" % cwd)

    if env is None:
        env = dict(os.environ)

    # This is just to ensure we get an exception if there is
    # something wrong with the pidfile
    if pidfile is not None:
        try:
            open(pidfile, "w").close()
        except Exception:
            raise EMError("Unable to write pidfile: %s" % pidfile)

    pid1 = os.fork()
    if pid1 == 0:
        if cwd is not None:
            os.chdir(cwd)

        pid2 = os.fork()
        if pid2 == 0:
            os.execve("/bin/sh", ["/bin/sh", "-c", script], env)
            os._exit(255)

        if pidfile is not None:
            f = open(pidfile, "w")
            f.write("%d\n" % pid2)
            f.close()

        os._exit(0)

    pid, exitcode = os.waitpid(pid1, 0)
    if exitcode != 0:
        raise EMError("Non-zero exitcode launching script: %d" % exitcode)


class WorkflowProcessor:
    def __init__(self, dao, workflow):
        self.dao = dao
        self.workflow = workflow

    def plan(self):
        "Launch the pegasus planner"
        w = self.workflow
        basedir = w.get_basedir()
        pidfile = w.get_pidfile()
        logfile = w.get_logfile()
        runfile = w.get_runfile()
        resultfile = w.get_resultfile()
        plan_command = w.get_plan_command()

        if os.path.isfile(pidfile) and self.planning():
            raise EMError("Planner already running")

        # When we re-plan, we need to remove all the old
        # files so that the ensemble manager doesn't get
        # confused.
        files = [runfile, resultfile, pidfile]
        for f in files:
            if os.path.isfile(f):
                os.remove(f)

        script = "({}) 2>&1 | tee -a {} | grep pegasus-run >{} ; /bin/echo $? >{}".format(
            plan_command, logfile, runfile, resultfile,
        )
        forkscript(script, cwd=basedir, pidfile=pidfile, env=get_script_env())

    def planning(self):
        "Check pidfile to see if the planner is still running"
        pidfile = self.workflow.get_pidfile()

        if not os.path.exists(pidfile):
            raise EMError("pidfile missing")

        pid = int(open(pidfile).read())

        try:
            os.kill(pid, 0)
            # If that succeeds, the process is still running
            return True
        except OSError as e:
            # errno 3 is No Such Process
            if e.errno != 3:
                raise

        return False

    def planning_successful(self):
        "Check to see if planning was successful"
        resultfile = self.workflow.get_resultfile()

        if not os.path.exists(resultfile):
            raise EMError("Result file not found: %s" % resultfile)

        exitcode = int(open(resultfile).read())

        if exitcode != 0:
            return False

        try:
            self.find_submitdir()
        except Exception as e:
            log.exception(e)
            return False

        return True

    def find_submitdir(self):
        "Get the workflow submitdir from the workflow log"
        logfile = self.workflow.get_runfile()

        if not os.path.isfile(logfile):
            raise EMError("Workflow run file not found: %s" % logfile)

        submitdir = None

        f = open(logfile)
        try:
            for l in f:
                if l.startswith("pegasus-run"):
                    submitdir = l.split()[1]
        finally:
            f.close()

        if submitdir is None:
            raise EMError("No pegasus-run found in the workflow run file: %s" % logfile)

        return submitdir

    def get_wf_uuid(self):
        "Get the workflow UUID from the braindump file"
        submitdir = self.find_submitdir()

        braindump_file_path = os.path.join(submitdir, "braindump.yml")

        if not os.path.isfile(braindump_file_path):
            raise EMError("braindump.yml not found")

        with open(braindump_file_path) as f:
            bd = braindump.load(f)

        if bd.wf_uuid is None:
            raise EMError("wf_uuid not found in braindump.yml")

        return bd.wf_uuid

    def run(self):
        "Run the workflow using pegasus-run"
        submitdir = self.workflow.submitdir

        if submitdir is None:
            raise EMError("Workflow submitdir not set")

        if not os.path.isdir(submitdir):
            raise EMError("Workflow submit dir does not exist: %s" % submitdir)

        logfile = self.workflow.get_logfile()

        runscript(
            "pegasus-run {} >>{} 2>&1".format(submitdir, logfile), env=get_script_env()
        )

    def get_dashboard(self):
        "Get the dashboard record for the workflow"
        wf_uuid = self.workflow.wf_uuid
        if wf_uuid is None:
            raise EMError("wf_uuid is none")

        try:
            w = (
                self.dao.session.query(MasterWorkflow)
                .filter_by(wf_uuid=str(wf_uuid))
                .one()
            )
            return w
        except NoResultFound:
            name = self.workflow.name
            log.debug("No dashboard record for workflow %s" % name)
            return None

    def get_dashboard_state_for_running_workflow(self):
        """Get the latest state of the workflow from the dashboard
        tables where timestamp is > last updated of the ensemble workflow"""
        # We can only use this for running workflows because we are assuming
        # that the last state of the workflow should be after the updated
        # timestamp of the ensemble workflow. That might not be true for
        # workflows in states other than RUNNING.
        if self.workflow.state != EnsembleWorkflowStates.RUNNING:
            raise EMError("This method should only be called for running workflows")

        w = self.get_dashboard()
        if w is None:
            raise EMError("Dashboard workflow not found")

        # Need to compute the unix ts for updated in this ugly way
        updated = (
            self.workflow.updated - datetime.datetime(1970, 1, 1)
        ).total_seconds()

        # Get the last event for the workflow where the event timestamp is
        # greater than the last updated ts for the ensemble workflow
        ws = (
            self.dao.session.query(MasterWorkflowstate)
            .filter_by(wf_id=w.wf_id)
            .filter(MasterWorkflowstate.timestamp >= updated)
            .order_by(text("timestamp desc"))
            .first()
        )

        if ws is None:
            name = self.workflow.name
            log.info("No recent workflow state records for workflow %s" % name)

        return ws

    def pending(self):
        "The workflow is pending if there is no dashboard record"
        w = self.get_dashboard()
        return w is None

    def running(self):
        "Is the workflow running"
        ws = self.get_dashboard_state_for_running_workflow()
        if ws is None:
            return True
        return ws.state == "WORKFLOW_STARTED"

    def running_successful(self):
        "Assuming the workflow is done running, did it finish successfully?"
        ws = self.get_dashboard_state_for_running_workflow()
        if ws is None or ws.state == "WORKFLOW_STARTED":
            raise EMError("Workflow is running")
        return ws.status == 0


class EnsembleProcessor:
    Processor = WorkflowProcessor

    def __init__(self, dao, ensemble):
        self.dao = dao
        self.ensemble = ensemble
        self.active = ensemble.state == EnsembleStates.ACTIVE
        self.max_running = ensemble.max_running
        self.max_planning = ensemble.max_planning
        self.running = 0
        self.planning = 0

        # We are only interested in workflows that are in one
        # of the following states. The EM doesn't handle other
        # states.
        active_states = {
            EnsembleWorkflowStates.READY,
            EnsembleWorkflowStates.PLANNING,
            EnsembleWorkflowStates.QUEUED,
            EnsembleWorkflowStates.RUNNING,
        }

        # We need a copy of the list so we can sort it
        self.workflows = [w for w in ensemble.workflows if w.state in active_states]

        # Sort workflows by priority
        def get_priority(w):
            return w.priority

        self.workflows.sort(key=get_priority, reverse=True)

        # Count the number of planning and running workflows
        for w in self.workflows:
            if w.state == EnsembleWorkflowStates.PLANNING:
                self.planning += 1
            elif w.state == EnsembleWorkflowStates.RUNNING:
                self.running += 1

    def can_plan(self):
        return self.active and self.planning < self.max_planning

    def plan_workflow(self, workflow):
        log.info("Planning %s" % workflow.name)
        self.planning += 1
        workflow.set_state(EnsembleWorkflowStates.PLANNING)
        workflow.set_updated()

        self.dao.session.flush()

        # Fork planning task
        p = self.Processor(self.dao, workflow)
        try:
            p.plan()
        except Exception as e:
            log.error("Planning failed for workflow %s" % workflow.name)
            log.exception(e)
            workflow.set_state(EnsembleWorkflowStates.PLAN_FAILED)
            workflow.set_updated()

        self.dao.session.commit()

    def handle_ready(self, workflow):
        if self.can_plan():
            self.plan_workflow(workflow)
        else:
            log.debug("Delaying planning of workflow %s due to policy" % workflow.name)

    def handle_planning(self, workflow):
        p = self.Processor(self.dao, workflow)

        if p.planning():
            log.info("Workflow %s is still planning" % workflow.name)
            return

        log.info("Workflow %s is no longer planning" % workflow.name)
        self.planning -= 1

        # Planning failed
        if not p.planning_successful():
            log.error("Planning failed for workflow %s" % workflow.name)
            workflow.set_state(EnsembleWorkflowStates.PLAN_FAILED)
            workflow.set_updated()
            self.dao.session.commit()
            return

        log.info("Queueing workflow %s" % workflow.name)

        # Planning succeeded, get uuid and queue workflow
        workflow.set_wf_uuid(p.get_wf_uuid())
        workflow.set_submitdir(p.find_submitdir())
        workflow.set_state(EnsembleWorkflowStates.QUEUED)
        workflow.set_updated()
        self.dao.session.commit()

        # Go ahead and handle the queued state now
        self.handle_queued(workflow)

    def can_run(self):
        return self.active and self.running < self.max_running

    def run_workflow(self, workflow):
        log.info("Running workflow %s" % workflow.name)

        self.running += 1
        workflow.set_state(EnsembleWorkflowStates.RUNNING)
        workflow.set_updated()
        self.dao.session.flush()

        p = self.Processor(self.dao, workflow)
        try:
            p.run()
        except Exception as e:
            log.debug("Running of workflow %s failed" % workflow.name)
            log.exception(e)
            workflow.set_state(EnsembleWorkflowStates.RUN_FAILED)
            workflow.set_updated()

        self.dao.session.commit()

    def handle_queued(self, workflow):
        if self.can_run():
            self.run_workflow(workflow)
        else:
            log.debug("Delaying run of workflow %s due to policy" % workflow.name)

    def handle_running(self, workflow):
        p = self.Processor(self.dao, workflow)

        if p.pending():
            log.info("Workflow %s is pending" % workflow.name)
            return

        if p.running():
            log.info("Workflow %s is running" % workflow.name)
            return

        log.info("Workflow %s is no longer running" % workflow.name)
        self.running -= 1

        if p.running_successful():
            log.info("Workflow %s was successful" % workflow.name)
            workflow.set_state(EnsembleWorkflowStates.SUCCESSFUL)
        else:
            log.info("Workflow %s failed" % workflow.name)
            workflow.set_state(EnsembleWorkflowStates.FAILED)

        workflow.set_updated()

        self.dao.session.commit()

    def handle_workflow(self, w):
        if w.state == EnsembleWorkflowStates.READY:
            self.handle_ready(w)
        elif w.state == EnsembleWorkflowStates.PLANNING:
            self.handle_planning(w)
        elif w.state == EnsembleWorkflowStates.QUEUED:
            self.handle_queued(w)
        elif w.state == EnsembleWorkflowStates.RUNNING:
            self.handle_running(w)

    def run(self):
        edir = self.ensemble.get_localdir()
        if not os.path.isdir(edir):
            log.info("Creating ensemble directory: %s" % edir)
            os.makedirs(edir, 0o700)

        log.info("Processing %d ensemble workflows..." % len(self.workflows))
        for w in self.workflows:
            try:
                self.handle_workflow(w)
            except Exception as e:
                log.error(
                    "Processing workflow %s of ensemble %s"
                    % (w.name, self.ensemble.name)
                )
                log.exception(e)
                self.dao.session.rollback()


class EnsembleManager(threading.Thread):
    Processor = EnsembleProcessor

    def __init__(self, interval=None):
        threading.Thread.__init__(self)
        self.daemon = True
        if interval is None:
            interval = float(emapp.config["EM_INTERVAL"])
        self.interval = interval

    def run(self):
        log.info("Ensemble Manager starting")
        self.loop_forever()

    def loop_forever(self):
        while True:
            u = user.get_user_by_uid(os.getuid())
            session = connection.connect(
                u.get_master_db_url(), connect_args={"check_same_thread": False}
            )
            try:
                dao = Ensembles(session)
                self.loop_once(dao)
            finally:
                session.close()
            time.sleep(self.interval)

    def loop_once(self, dao):
        actionable = dao.list_actionable_ensembles()
        if len(actionable) == 0:
            return

        log.info("Processing ensembles")
        for e in actionable:
            log.info("Processing ensemble %s", e.name)
            p = self.Processor(dao, e)
            p.run()



#!/usr/bin/env python3

import logging
import optparse
import os
import re
import sys
import traceback

from Pegasus.db import connection
from Pegasus.db.workflow.stampede_statistics import StampedeStatistics
from Pegasus.db.workflow.stampede_wf_statistics import StampedeWorkflowStatistics
from Pegasus.plots_stats import utils as stats_utils
from Pegasus.tools import utils

root_logger = logging.getLogger()
logger = logging.getLogger("pegasus-statistics")


utils.configureLogging(level=logging.WARNING)

# Regular expressions
re_parse_property = re.compile(r"([^:= \t]+)\s*[:=]?\s*(.*)")

workflow_summary_file_name = "summary"
workflow_summary_time_file_name = "summary-time"
workflow_statistics_file_name = "workflow"
job_statistics_file_name = "jobs"
logical_transformation_statistics_file_name = "breakdown"
logical_integrity_statistics_file_name = "integrity"
time_statistics_file_name = "time"
time_statistics_per_host_file_name = "time-per-host"
text_file_extension = ".txt"
csv_file_extension = ".csv"
calc_wf_stats = False
calc_wf_summary = False
calc_jb_stats = False
calc_tf_stats = False
calc_ti_stats = False
calc_int_stats = False
time_filter = None
NEW_LINE_STR = "\n"
DEFAULT_OUTPUT_DIR = "statistics"
FILE_TYPE_TXT = "text"
FILE_TYPE_CSV = "csv"
uses_PMC = False

# Transformations file column names
transformation_stats_col_name_text = [
    "Transformation",
    "Type",
    "Count",
    "Min (runtime)",
    "Max (runtime)",
    "Mean (runtime)",
    "Total (runtime)",
    "Min (mem)",
    "Max (mem)",
    "Mean (mem)",
    "Min (avg. cpu)",
    "Max (avg. cpu)",
    "Mean (avg. cpu)",
]
transformation_stats_col_name_csv = [
    "Workflow_Id",
    "Dax_Label",
    "Transformation",
    "Type",
    "Count",
    "Min (runtime)",
    "Max (runtime)",
    "Mean (runtime)",
    "Total (runtime)",
    "Min (mem)",
    "Max (mem)",
    "Mean (mem)",
    "Min (avg. cpu)",
    "Max (avg. cpu)",
    "Mean (avg. cpu)",
]

transformation_stats_col_size = [
    15,
    5,
    6,
    14,
    14,
    21,
    22,
    10,
    10,
    11,
    15,
    15,
    16,
]

# Integrity file column names
integrity_stats_col_name_text = ["Type", "File Type", "Count", "Total Duration"]
integrity_stats_col_name_csv = [
    "Workflow_Id",
    "Dax_Label",
    "Type",
    "File Type",
    "Count",
    "Total Duration",
]
integrity_stats_col_size = [10, 15, 10, 25]

# Jobs file column names
job_stats_col_name_text = [
    "Job",
    "Try",
    "Site",
    "Kickstart",
    "Mult",
    "Kickstart-Mult",
    "CPU-Time",
    "Post",
    "CondorQTime",
    "Resource",
    "Runtime",
    "Seqexec",
    "Seqexec-Delay",
    "Exitcode",
    "Hostname",
]
job_stats_col_name_csv = [
    "Workflow_Id",
    "Dax_Label",
    "Job",
    "Try",
    "Site",
    "Kickstart",
    "Mult",
    "Kickstart-Mult",
    "CPU-Time",
    "Post",
    "CondorQTime",
    "Resource",
    "Runtime",
    "Seqexec",
    "Seqexec-Delay",
    "Exitcode",
    "Hostname",
]
job_stats_col_size = [35, 4, 12, 12, 6, 16, 12, 6, 12, 12, 12, 12, 15, 10, 30]

# Summary file column names
workflow_summary_col_name_csv = [
    "Type",
    "Succeeded",
    "Failed",
    "Incomplete",
    "Total",
    "Retries",
    "Total+Retries",
]
workflow_summary_col_name_text = [
    "Type",
    "Succeeded",
    "Failed",
    "Incomplete",
    "Total",
    "Retries",
    "Total+Retries",
]
workflow_summary_col_size = [15, 10, 8, 12, 10, 10, 13]
workflow_time_summary_col_name_csv = ["stat_type", "time_seconds"]

# Workflow file column names
workflow_status_col_name_text = [
    "Type",
    "Succeeded",
    "Failed",
    "Incomplete",
    "Total",
    "Retries",
    "Total+Retries",
    "WF Retries",
]
workflow_status_col_name_csv = [
    "Workflow_Id",
    "Dax_Label",
    "Type",
    "Succeeded",
    "Failed",
    "Incomplete",
    "Total",
    "Retries",
    "Total+Retries",
    "WF_Retries",
]
workflow_status_col_size = [15, 11, 10, 12, 10, 10, 15, 10]

# Time file column names
time_stats_col_name_csv = ["stat_type", "date", "count", "runtime (sec)"]
time_stats_col_name_text = ["Date", "Count", "Runtime (sec)"]
time_stats_col_size = [30, 20, 20]
time_host_stats_col_name_csv = ["stat_type", "date", "host", "count", "runtime (sec)"]
time_host_stats_col_name_text = ["Date", "Host", "Count", "Runtime (sec)"]
time_host_stats_col_size = [23, 30, 10, 20]


class JobStatistics:
    def __init__(self):
        self.name = None
        self.site = None
        self.kickstart = None
        self.multiplier_factor = None
        self.kickstart_mult = None
        self.remote_cpu_time = None
        self.post = None
        self.condor_delay = None
        self.resource = None
        self.runtime = None
        self.condorQlen = None
        self.seqexec = None
        self.seqexec_delay = None
        self.retry_count = 0
        self.exitcode = None
        self.hostname = None

    def getFormattedJobStatistics(self):
        return [
            self.name,
            str(self.retry_count),
            self.site or "-",
            fstr(self.kickstart),
            str(self.multiplier_factor),
            fstr(self.kickstart_mult),
            fstr(self.remote_cpu_time),
            fstr(self.post),
            fstr(self.condor_delay),
            fstr(self.resource),
            fstr(self.runtime),
            fstr(self.seqexec),
            fstr(self.seqexec_delay),
            str(self.exitcode),
            self.hostname,
        ]


def formatted_wf_summary_legends_part1():
    return """
#
# Pegasus Workflow Management System - http://pegasus.isi.edu
#
# Workflow summary:
#   Summary of the workflow execution. It shows total
#   tasks/jobs/sub workflows run, how many succeeded/failed etc.
#   In case of hierarchical workflow the calculation shows the
#   statistics across all the sub workflows.It shows the following
#   statistics about tasks, jobs and sub workflows.
#     * Succeeded - total count of succeeded tasks/jobs/sub workflows.
#     * Failed - total count of failed tasks/jobs/sub workflows.
#     * Incomplete - total count of tasks/jobs/sub workflows that are
#       not in succeeded or failed state. This includes all the jobs
#       that are not submitted, submitted but not completed etc. This
#       is calculated as  difference between 'total' count and sum of
#       'succeeded' and 'failed' count.
#     * Total - total count of tasks/jobs/sub workflows.
#     * Retries - total retry count of tasks/jobs/sub workflows.
#     * Total+Retries - total count of tasks/jobs/sub workflows executed
#       during workflow run. This is the cumulative of retries,
#       succeeded and failed count."""


def formatted_wf_summary_legends_part2():
    return """
# Workflow wall time:
#   The wall time from the start of the workflow execution to the end as
#   reported by the DAGMAN.In case of rescue dag the value is the
#   cumulative of all retries.
# Cumulative job wall time:
#   The sum of the wall time of all jobs as reported by kickstart.
#   In case of job retries the value is the cumulative of all retries.
#   For workflows having sub workflow jobs (i.e SUBDAG and SUBDAX jobs),
#   the wall time value includes jobs from the sub workflows as well.
# Cumulative job wall time as seen from submit side:
#   The sum of the wall time of all jobs as reported by DAGMan.
#   This is similar to the regular cumulative job wall time, but includes
#   job management overhead and delays. In case of job retries the value
#   is the cumulative of all retries. For workflows having sub workflow
#   jobs (i.e SUBDAG and SUBDAX jobs), the wall time value includes jobs
#   from the sub workflows as well.
# Cumulative job badput wall time:
#   The sum of the wall time of all failed jobs as reported by kickstart.
#   In case of job retries the value is the cumulative of all retries.
#   For workflows having sub workflow jobs (i.e SUBDAG and SUBDAX jobs),
#   the wall time value includes jobs from the sub workflows as well.
# Cumulative job badput wall time as seen from submit side:
#   The sum of the wall time of all failed jobs as reported by DAGMan.
#   This is similar to the regular cumulative job badput wall time, but includes
#   job management overhead and delays. In case of job retries the value
#   is the cumulative of all retries. For workflows having sub workflow
#   jobs (i.e SUBDAG and SUBDAX jobs), the wall time value includes jobs
#   from the sub workflows as well."""


def formatted_wf_summary_legends_txt():
    return formatted_wf_summary_legends_part1() + formatted_wf_summary_legends_part2()


def formatted_wf_summary_legends_csv1():
    return formatted_wf_summary_legends_part1()


def formatted_wf_summary_legends_csv2():
    return formatted_wf_summary_legends_part2()


def formatted_wf_status_legends():
    return """
# Workflow summary
#   Summary of the workflow execution. It shows total
#   tasks/jobs/sub workflows run, how many succeeded/failed etc.
#   In case of hierarchical workflow the calculation shows the
#   statistics of each individual sub workflow.The file also
#   contains a 'Total' table at the bottom which is the cummulative
#   of all the individual statistics details.t shows the following
#   statistics about tasks, jobs and sub workflows.
#
#     * WF Retries - number of times a workflow was retried.
#     * Succeeded - total count of succeeded tasks/jobs/sub workflows.
#     * Failed - total count of failed tasks/jobs/sub workflows.
#     * Incomplete - total count of tasks/jobs/sub workflows that are
#       not in succeeded or failed state. This includes all the jobs
#       that are not submitted, submitted but not completed etc. This
#       is calculated as  difference between 'total' count and sum of
#       'succeeded' and 'failed' count.
#     * Total - total count of tasks/jobs/sub workflows.
#     * Retries - total retry count of tasks/jobs/sub workflows.
#     * Total+Retries - total count of tasks/jobs/sub workflows executed
#       during workflow run. This is the cumulative of retries,
#       succeeded and failed count.
#
"""


def formatted_job_stats_legends():
    return """
# Job            - name of the job
# Try            - number representing the job instance run count
# Site           - site where the job ran
# Kickstart      - actual duration of the job instance in seconds on the
#                  remote compute node
# Mult           - multiplier factor specified by the user
# Kickstart-Mult - Kickstart time multiplied by the multiplier factor
# CPU-Time       - remote cpu time computed as the stime + utime
# Post           - postscript time as reported by DAGMan
# CondorQTime    - time between submission by DAGMan and the remote Grid
#                  submission. It is an estimate of the time spent in the
#                  condor q on the submit node
# Resource       - time between the remote Grid submission and start of
#                  remote execution. It is an estimate of the time job
#                  spent in the remote queue
# Runtime        - time spent on the resource as seen by Condor DAGMan.
#                  Is always >= Kickstart
# Seqexec        - time taken for the completion of a clustered job
# Seqexec-Delay  - time difference between the time for the completion
#                  of a clustered job and sum of all the individual
#                  tasks Kickstart time
# Exitcode       - exitcode for this job
# Hostname       - name of the host where the job ran, as reported by
#                  Kickstart"""


def formatted_transformation_stats_legends():
    return """
# Transformation   - name of the transformation.
# Type             - successful or failed
# Count            - the number of times the invocations corresponding to
#                    the transformation was executed.
# Min(sec)         - the minimum invocation runtime value corresponding
#                    to the transformation.
# Max(sec)         - the maximum invocation runtime value corresponding
#                    to the transformation.
# Mean(sec)        - the mean of the invocation runtime corresponding
#                    to the transformation.
# Total(sec)       - the cumulative of invocation runtime corresponding
#                    to the transformation.
# Min (mem)        - the minimum of the max. resident set size (RSS) value corresponding
#                    to the transformation. In MB.
# Max (mem)        - the maximum of the max. resident set size (RSS) value corresponding
#                    to the transformation. In MB.
# Mean (mem)       - the mean of the max. resident set size (RSS) value corresponding
#                    to the transformation. In MB.
# Min (avg. cpu)   - the minimum of the average cpu utilization value corresponding
#                    to the transformation.
# Max (avg. cpu)   - the maximum of the average cpu utilization value corresponding
#                    to the transformation.
# Mean (avg. cpu)  - the mean of the average cpu utilization value corresponding
#                    to the transformation."""


def formatted_integrity_stats_legends():
    return """
# Integrity Statistics Breakdown
# Type           - the type of integrity metric.
                   check - means checksum was compared for a file,
                   compute - means a checksum was generated for a file
# File Type      - the type of file: input or output from a job perspective
# Count          - the number of times done
# Total Duration - sum of duration in seconds for the 'count' number
                   of records matching the particular type,
                   file-type combo
"""


def formatted_time_stats_legends_text():
    return """
# Job instance statistics per FILTER:
#   the number of job instances run, total runtime sorted by FILTER
# Invocation statistics per FILTER:
#   the number of invocations , total runtime sorted by FILTER
# Job instance statistics by host per FILTER:
#   the number of job instance run, total runtime on each host sorted by FILTER
# Invocation by host per FILTER:
#   the number of invocations, total runtime on each host sorted by FILTER
""".replace(
        "FILTER", str(time_filter)
    )


def formatted_time_stats_legends_csv():
    return """
# Job instance statistics per FILTER:
#   the number of job instances run, total runtime sorted by FILTER
# Invocation statistics per FILTER:
#   the number of invocations , total runtime sorted by FILTER
""".replace(
        "FILTER", str(time_filter)
    )


def formatted_time_host_stats_legends_csv():
    return """
# Job instance statistics by host per FILTER:
#   the number of job instance run, total runtime on each host sorted by FILTER
# Invocation by host per FILTER:
#   the number of invocations, total runtime on each host sorted by FILTER
""".replace(
        "FILTER", str(time_filter)
    )


def remove_file(path):
    try:
        os.remove(path)
    except OSError as e:
        if e.errno != 2:
            raise


def write_to_file(file_path, mode, content):
    """
    Utility method for writing content to a given file
    @param file_path :  file path
    @param mode :   file writing mode 'a' append , 'w' write
    @param content :  content to write to file
    """
    try:
        fh = open(file_path, mode)
        fh.write(content)
    except OSError:
        logger.error("Unable to write to file " + file_path)
        sys.exit(1)
    else:
        fh.close()


def format_seconds(duration):
    """
    Utility for converting time to a readable format
    @param duration :  time in seconds and miliseconds
    @return time in format day,hour, min,sec
    """
    return stats_utils.format_seconds(duration)


def istr(value):
    """
    Utility for returning a str representation of the given value.
    Return '-' if value is None
    @parem value : the given value that need to be converted to string
    """
    if value is None:
        return "-"
    return str(value)


def fstr(value, to=3):
    """
    Utility method for rounding the float value to rounded string
    @param value :  value to round
    @param to    :  how many decimal points to round to
    """
    if value is None:
        return "-"
    return stats_utils.round_decimal_to_str(value, to)


def pstr(value, to=2):
    """
    Utility method for rounding the float value to rounded string
    @param value :  value to round
    @param to    :  how many decimal points to round to
    """
    if value is None:
        return "-"
    return stats_utils.round_decimal_to_str(value, to) + "%"


def print_row(row, sizes, fmt):
    """
    Utility method for generating formatted row based on the column format given
    row   : list of column values
    sizes : list of column widths for text format
    fmt   : format of the row ('text' or 'csv')
    """
    if fmt in ["text", "txt"]:
        return "".join(value.ljust(sizes[i]) for i, value in enumerate(row))
    elif fmt == "csv":
        return ",".join(row)
    else:
        print("Output format %s not recognized!" % fmt)
        sys.exit(1)


def print_workflow_details(output_db_url, wf_uuid, output_dir, multiple_wf=False):
    """
    Prints the workflow statistics information of all workflows
    @param output_db_url : URL of stampede DB
    @param wf_uuid       : uuid of the top level workflow
    @param output_dir    : directory to write output files
    """
    errors = 0
    try:
        if multiple_wf:
            expanded_workflow_stats = StampedeWorkflowStatistics(output_db_url)
        else:
            expanded_workflow_stats = StampedeStatistics(output_db_url)

        wf_found = expanded_workflow_stats.initialize(wf_uuid)

        if wf_found is False:
            print(
                "Workflow {!r} not found in database {!r}".format(
                    wf_uuid, output_db_url
                )
            )
            sys.exit(1)
    except Exception:
        logger.error("Failed to load the database." + output_db_url)
        logger.warning(traceback.format_exc())
        sys.exit(1)

    # print workflow statistics
    if multiple_wf:
        wf_uuid_list = []
        desc_wf_uuid_list = expanded_workflow_stats.get_workflow_ids()
    else:
        wf_uuid_list = [wf_uuid]
        desc_wf_uuid_list = expanded_workflow_stats.get_descendant_workflow_ids()

    for wf_det in desc_wf_uuid_list:
        wf_uuid_list.append(wf_det.wf_uuid)

    global calc_wf_stats
    if calc_wf_stats:
        if file_type == FILE_TYPE_TXT:
            wf_stats_file_txt = os.path.join(
                output_dir, workflow_statistics_file_name + text_file_extension
            )
            write_to_file(wf_stats_file_txt, "w", formatted_wf_status_legends())
            header = print_row(
                workflow_status_col_name_text, workflow_status_col_size, "text"
            )
            write_to_file(wf_stats_file_txt, "a", header)

        if file_type == FILE_TYPE_CSV:
            wf_stats_file_csv = os.path.join(
                output_dir, workflow_statistics_file_name + csv_file_extension
            )
            write_to_file(wf_stats_file_csv, "w", formatted_wf_status_legends())
            header = print_row(
                workflow_status_col_name_csv, workflow_status_col_size, "csv"
            )
            write_to_file(wf_stats_file_csv, "a", header)

    global calc_jb_stats
    if calc_jb_stats:
        jobs_stats_file_txt = os.path.join(
            output_dir, job_statistics_file_name + text_file_extension
        )
        if file_type == FILE_TYPE_TXT:
            write_to_file(jobs_stats_file_txt, "w", formatted_job_stats_legends())

        jobs_stats_file_csv = os.path.join(
            output_dir, job_statistics_file_name + csv_file_extension
        )
        if file_type == FILE_TYPE_CSV:
            write_to_file(jobs_stats_file_csv, "w", formatted_job_stats_legends())

    global calc_tf_stats
    if calc_tf_stats:
        if file_type == FILE_TYPE_TXT:
            transformation_stats_file_txt = os.path.join(
                output_dir,
                logical_transformation_statistics_file_name + text_file_extension,
            )
            write_to_file(
                transformation_stats_file_txt,
                "w",
                formatted_transformation_stats_legends(),
            )

        if file_type == FILE_TYPE_CSV:
            transformation_stats_file_csv = os.path.join(
                output_dir,
                logical_transformation_statistics_file_name + csv_file_extension,
            )
            write_to_file(
                transformation_stats_file_csv,
                "w",
                formatted_transformation_stats_legends(),
            )

    global calc_int_stats
    if calc_int_stats:
        if file_type == FILE_TYPE_TXT:
            integrity_stats_file_txt = os.path.join(
                output_dir, logical_integrity_statistics_file_name + text_file_extension
            )
            write_to_file(
                integrity_stats_file_txt, "w", formatted_integrity_stats_legends()
            )

        if file_type == FILE_TYPE_CSV:
            integrity_stats_file_csv = os.path.join(
                output_dir, logical_integrity_statistics_file_name + csv_file_extension
            )
            write_to_file(
                integrity_stats_file_csv, "w", formatted_integrity_stats_legends()
            )

    global calc_ti_stats
    if calc_ti_stats:
        try:
            time_stats_file_txt = os.path.join(
                output_dir, time_statistics_file_name + text_file_extension
            )
            if file_type == FILE_TYPE_TXT:
                content = print_statistics_by_time_and_host(
                    expanded_workflow_stats, "text", combined=True, per_host=True
                )
                write_to_file(
                    time_stats_file_txt, "w", formatted_time_stats_legends_text()
                )
                write_to_file(time_stats_file_txt, "a", content)

            time_stats_file_csv = os.path.join(
                output_dir, time_statistics_file_name + csv_file_extension
            )
            if file_type == FILE_TYPE_CSV:
                content = print_statistics_by_time_and_host(
                    expanded_workflow_stats, "csv", combined=True, per_host=False
                )
                write_to_file(
                    time_stats_file_csv, "w", formatted_time_stats_legends_csv()
                )
                write_to_file(time_stats_file_csv, "a", content)

                time_stats_file2_csv = os.path.join(
                    output_dir, time_statistics_per_host_file_name + csv_file_extension
                )
                content = print_statistics_by_time_and_host(
                    expanded_workflow_stats, "csv", combined=False, per_host=True
                )
                write_to_file(
                    time_stats_file2_csv, "w", formatted_time_host_stats_legends_csv()
                )
                write_to_file(time_stats_file2_csv, "a", content)
        except Exception:
            logger.warn("time statistics generation failed")
            logger.debug("time statistics generation failed", exc_info=1)

            calc_ti_stats = False
            if file_type == FILE_TYPE_TXT:
                remove_file(time_stats_file_txt)
            if file_type == FILE_TYPE_CSV:
                remove_file(time_stats_file_csv)
                remove_file(time_stats_file2_csv)

            errors += 1

    if calc_jb_stats or calc_tf_stats or calc_wf_stats or calc_int_stats:
        for sub_wf_uuid in wf_uuid_list:
            try:
                individual_workflow_stats = StampedeStatistics(output_db_url, False)
                wf_found = individual_workflow_stats.initialize(sub_wf_uuid)

                if wf_found is False:
                    print(
                        "Workflow %r not found in database %r"
                        % (sub_wf_uuid, output_db_url)
                    )
                    sys.exit(1)
            except Exception:
                logger.error("Failed to load the database." + output_db_url)
                logger.warning(traceback.format_exc())
                sys.exit(1)

            wf_det = individual_workflow_stats.get_workflow_details()[0]

            workflow_id = str(sub_wf_uuid)
            dax_label = str(wf_det.dax_label)
            logger.info(
                "Generating statistics information about the workflow "
                + workflow_id
                + " ... "
            )

            if calc_jb_stats:
                try:
                    logger.debug(
                        "Generating job instance statistics information for workflow "
                        + workflow_id
                        + " ... "
                    )
                    individual_workflow_stats.set_job_filter("all")

                    if file_type == FILE_TYPE_TXT:
                        content = print_individual_wf_job_stats(
                            individual_workflow_stats, workflow_id, dax_label, "text"
                        )
                        write_to_file(jobs_stats_file_txt, "a", content)

                    if file_type == FILE_TYPE_CSV:
                        content = print_individual_wf_job_stats(
                            individual_workflow_stats, workflow_id, dax_label, "csv"
                        )
                        write_to_file(jobs_stats_file_csv, "a", content)
                except Exception:
                    logger.warn("job instance statistics generation failed")
                    logger.debug(
                        "job instance statistics generation failed", exc_info=1
                    )

                    calc_jb_stats = False
                    if file_type == FILE_TYPE_TXT:
                        remove_file(jobs_stats_file_txt)
                    if file_type == FILE_TYPE_CSV:
                        remove_file(jobs_stats_file_csv)

                    errors += 1

            if calc_tf_stats:
                try:
                    logger.debug(
                        "Generating invocation statistics information for workflow "
                        + workflow_id
                        + " ... "
                    )
                    individual_workflow_stats.set_job_filter("all")
                    if file_type == FILE_TYPE_TXT:
                        content = print_wf_transformation_stats(
                            individual_workflow_stats, workflow_id, dax_label, "text"
                        )
                        write_to_file(transformation_stats_file_txt, "a", content)

                    if file_type == FILE_TYPE_CSV:
                        content = print_wf_transformation_stats(
                            individual_workflow_stats, workflow_id, dax_label, "csv"
                        )
                        write_to_file(transformation_stats_file_csv, "a", content)
                except Exception:
                    logger.warn("transformation statistics generation failed")
                    logger.debug(
                        "transformation statistics generation failed", exc_info=1
                    )

                    calc_tf_stats = False
                    if file_type == FILE_TYPE_TXT:
                        remove_file(transformation_stats_file_txt)
                    if file_type == FILE_TYPE_CSV:
                        remove_file(transformation_stats_file_csv)

                    errors += 1

            if calc_int_stats:
                try:
                    logger.debug(
                        "Generating integrity statistics information for workflow "
                        + workflow_id
                        + " ... "
                    )
                    individual_workflow_stats.set_job_filter("all")
                    if file_type == FILE_TYPE_TXT:
                        content = print_wf_integrity_stats(
                            individual_workflow_stats, workflow_id, dax_label, "text"
                        )
                        write_to_file(integrity_stats_file_txt, "a", content)

                    if file_type == FILE_TYPE_CSV:
                        content = print_wf_integrity_stats(
                            individual_workflow_stats, workflow_id, dax_label, "csv"
                        )
                        write_to_file(integrity_stats_file_csv, "a", content)
                except Exception:
                    logger.error(traceback.format_exc())
                    logger.warn("integrity statistics generation failed")
                    logger.debug("integrity statistics generation failed", exc_info=1)

                    calc_int_stats = False
                    if file_type == FILE_TYPE_TXT:
                        remove_file(integrity_stats_file_txt)
                    if file_type == FILE_TYPE_CSV:
                        remove_file(integrity_stats_file_csv)

                    errors += 1

            if calc_wf_stats:
                try:
                    logger.debug(
                        "Generating workflow statistics information for workflow "
                        + workflow_id
                        + " ... "
                    )
                    individual_workflow_stats.set_job_filter("all")
                    if file_type == FILE_TYPE_TXT:
                        content = print_individual_workflow_stats(
                            individual_workflow_stats, workflow_id, dax_label, "text"
                        )
                        write_to_file(wf_stats_file_txt, "a", content)

                    if file_type == FILE_TYPE_CSV:
                        content = print_individual_workflow_stats(
                            individual_workflow_stats, workflow_id, dax_label, "csv"
                        )
                        write_to_file(wf_stats_file_csv, "a", content)
                except Exception:
                    logger.warn("workflow statistics generation failed")
                    logger.debug("workflow statistics generation failed", exc_info=1)

                    calc_wf_stats = False
                    if file_type == FILE_TYPE_TXT:
                        remove_file(wf_stats_file_txt)
                    if file_type == FILE_TYPE_CSV:
                        remove_file(wf_stats_file_csv)

                    errors += 1

            individual_workflow_stats.close()

    #
    # Output printed to console
    #

    stats_output = ""

    global calc_wf_summary
    if calc_wf_summary:
        logger.info("Generating workflow summary ... ")
        try:
            if file_type == FILE_TYPE_TXT:
                summary_output = formatted_wf_summary_legends_txt()
                summary_output += NEW_LINE_STR
                summary_output += print_workflow_summary(
                    expanded_workflow_stats,
                    "text",
                    wf_summary=True,
                    time_summary=True,
                    multiple_wf=multiple_wf,
                )
                wf_summary_file_txt = os.path.join(
                    output_dir, workflow_summary_file_name + text_file_extension
                )
                write_to_file(wf_summary_file_txt, "w", summary_output)

                stats_output += summary_output + "\n"
                stats_output += "{:<30}: {}\n".format("Summary", wf_summary_file_txt)

            if file_type == FILE_TYPE_CSV:
                # Generate the first csv summary file
                summary_output = formatted_wf_summary_legends_csv1()
                summary_output += NEW_LINE_STR
                summary_output += print_workflow_summary(
                    expanded_workflow_stats,
                    "csv",
                    wf_summary=True,
                    time_summary=False,
                    multiple_wf=multiple_wf,
                )
                wf_summary_file_csv = os.path.join(
                    output_dir, workflow_summary_file_name + csv_file_extension
                )
                write_to_file(wf_summary_file_csv, "w", summary_output)

                stats_output += "{:<30}: {}\n".format("Summary:", wf_summary_file_csv)

                # Generate the second csv summary file
                summary_output = formatted_wf_summary_legends_csv2()
                summary_output += NEW_LINE_STR
                summary_output += print_workflow_summary(
                    expanded_workflow_stats,
                    "csv",
                    wf_summary=False,
                    time_summary=True,
                    multiple_wf=multiple_wf,
                )
                wf_summary_file2_csv = os.path.join(
                    output_dir, workflow_summary_time_file_name + csv_file_extension
                )
                write_to_file(wf_summary_file2_csv, "w", summary_output)

                stats_output += "{:<30}: {}\n".format(
                    "Summary Time:", wf_summary_file2_csv
                )
        except Exception:
            logger.warn("summary statistics generation failed")
            logger.debug("summary statistics generation failed", exc_info=1)
            stats_output = ""

            calc_wf_summary = False
            if file_type == FILE_TYPE_TXT:
                remove_file(wf_summary_file_txt)
            if file_type == FILE_TYPE_CSV:
                remove_file(workflow_summary_file_name)
                remove_file(wf_summary_file2_csv)

            errors += 1

    if calc_wf_stats:
        pre_stats_output = stats_output
        try:
            stats_output += "%-30s: " % "Workflow execution statistics"

            if file_type == FILE_TYPE_TXT:
                content = print_individual_workflow_stats(
                    expanded_workflow_stats, "All Workflows", "", "text"
                )
                write_to_file(wf_stats_file_txt, "a", content)
                stats_output += wf_stats_file_txt + "\n"

            if file_type == FILE_TYPE_CSV:
                content = print_individual_workflow_stats(
                    expanded_workflow_stats, "ALL", "", "csv"
                )
                write_to_file(wf_stats_file_csv, "a", content)
                stats_output += wf_stats_file_csv + "\n"
        except Exception:
            logger.warn("workflow statistics generation failed")
            logger.debug("workflow statistics generation failed", exc_info=1)
            stats_output = pre_stats_output

            calc_wf_stats = False
            if file_type == FILE_TYPE_TXT:
                remove_file(wf_stats_file_txt)
            if file_type == FILE_TYPE_CSV:
                remove_file(wf_stats_file_csv)

            errors += 1

    if calc_jb_stats:
        stats_output += "%-30s: " % "Job instance statistics"
        if file_type == FILE_TYPE_TXT:
            stats_output += jobs_stats_file_txt + "\n"

        if file_type == FILE_TYPE_CSV:
            stats_output += jobs_stats_file_csv + "\n"

    if calc_tf_stats:
        expanded_workflow_stats.set_job_filter("all")
        pre_stats_output = stats_output

        try:
            stats_output += "%-30s: " % "Transformation statistics"

            if file_type == FILE_TYPE_TXT:
                content = print_wf_transformation_stats(
                    expanded_workflow_stats, "All", "", "text"
                )
                write_to_file(transformation_stats_file_txt, "a", content)
                stats_output += transformation_stats_file_txt + "\n"

            if file_type == FILE_TYPE_CSV:
                content = print_wf_transformation_stats(
                    expanded_workflow_stats, "ALL", "", "csv"
                )
                write_to_file(transformation_stats_file_csv, "a", content)
                stats_output += transformation_stats_file_csv + "\n"
        except Exception:
            logger.warn("transformation statistics generation failed")
            logger.debug("transformation statistics generation failed", exc_info=1)
            stats_output = pre_stats_output

            calc_tf_stats = False
            if file_type == FILE_TYPE_TXT:
                remove_file(transformation_stats_file_txt)
            if file_type == FILE_TYPE_CSV:
                remove_file(transformation_stats_file_csv)
            errors += 1

    if calc_int_stats:
        expanded_workflow_stats.set_job_filter("all")
        pre_stats_output = stats_output

        try:
            stats_output += "%-30s: " % "Integrity statistics"

            if file_type == FILE_TYPE_TXT:
                content = print_wf_integrity_stats(
                    expanded_workflow_stats, "All", "", "text"
                )
                write_to_file(integrity_stats_file_txt, "a", content)
                stats_output += integrity_stats_file_txt + "\n"

            if file_type == FILE_TYPE_CSV:
                content = print_wf_integrity_stats(
                    expanded_workflow_stats, "ALL", "", "csv"
                )
                write_to_file(integrity_stats_file_csv, "a", content)
                stats_output += integrity_stats_file_csv + "\n"
        except Exception:
            logger.warn("integrity statistics generation failed")
            logger.debug("integrity statistics generation failed", exc_info=1)
            stats_output = pre_stats_output

            calc_int_stats = False
            if file_type == FILE_TYPE_TXT:
                remove_file(integrity_stats_file_txt)
            if file_type == FILE_TYPE_CSV:
                remove_file(integrity_stats_file_csv)

            errors += 1

    if calc_ti_stats:
        stats_output += "%-30s: " % "Time statistics"

        if file_type == FILE_TYPE_TXT:
            stats_output += time_stats_file_txt + "\n"

        if file_type == FILE_TYPE_CSV:
            stats_output += time_stats_file_csv + "\n"

    expanded_workflow_stats.close()

    print(stats_output)
    return errors


def print_workflow_summary(
    workflow_stats, output_format, wf_summary=True, time_summary=True, multiple_wf=False
):
    """
    Prints the workflow statistics summary of an top level workflow
    @param workflow_stats :  workflow statistics object reference
    """

    summary_str = ""

    if wf_summary is True:
        # status
        workflow_stats.set_job_filter("nonsub")

        # Tasks
        total_tasks = workflow_stats.get_total_tasks_status()
        total_succeeded_tasks = workflow_stats.get_total_succeeded_tasks_status(
            uses_PMC
        )
        total_failed_tasks = workflow_stats.get_total_failed_tasks_status()
        total_unsubmitted_tasks = total_tasks - (
            total_succeeded_tasks + total_failed_tasks
        )
        total_task_retries = workflow_stats.get_total_tasks_retries()
        total_invocations = (
            total_succeeded_tasks + total_failed_tasks + total_task_retries
        )

        # Jobs
        total_jobs = workflow_stats.get_total_jobs_status()
        total_succeeded_failed_jobs = (
            workflow_stats.get_total_succeeded_failed_jobs_status()
        )
        total_failed_jobs_integrity = workflow_stats.get_total_succeeded_failed_jobs_status(
            classify_error=True, tag="int.error"
        )
        total_succeeded_jobs = total_succeeded_failed_jobs.succeeded
        total_failed_jobs = total_succeeded_failed_jobs.failed
        total_unsubmitted_jobs = total_jobs - (total_succeeded_jobs + total_failed_jobs)
        total_job_retries = workflow_stats.get_total_jobs_retries()
        total_job_instance_retries = (
            total_succeeded_jobs + total_failed_jobs + total_job_retries
        )

        # Sub workflows
        workflow_stats.set_job_filter("subwf")
        total_sub_wfs = workflow_stats.get_total_jobs_status()
        total_succeeded_failed_sub_wfs = (
            workflow_stats.get_total_succeeded_failed_jobs_status()
        )
        total_succeeded_sub_wfs = total_succeeded_failed_sub_wfs.succeeded
        total_failed_sub_wfs = total_succeeded_failed_sub_wfs.failed
        # for non hierarichal workflows the combined query can return none
        if total_succeeded_sub_wfs is None:
            total_succeeded_sub_wfs = 0

        if total_failed_sub_wfs is None:
            total_failed_sub_wfs = 0

        total_unsubmitted_sub_wfs = total_sub_wfs - (
            total_succeeded_sub_wfs + total_failed_sub_wfs
        )
        total_sub_wfs_retries = workflow_stats.get_total_jobs_retries()
        total_sub_wfs_tries = (
            total_succeeded_sub_wfs + total_failed_sub_wfs + total_sub_wfs_retries
        )

        # Format the output
        if output_format == "text":
            summary_str += "".center(sum(workflow_summary_col_size), "-") + "\n"
            summary_str += (
                print_row(
                    workflow_summary_col_name_text,
                    workflow_summary_col_size,
                    output_format,
                )
                + "\n"
            )
        else:
            summary_str += (
                print_row(
                    workflow_summary_col_name_csv,
                    workflow_summary_col_size,
                    output_format,
                )
                + "\n"
            )

        content = [
            "Tasks",
            istr(total_succeeded_tasks),
            istr(total_failed_tasks),
            istr(total_unsubmitted_tasks),
            istr(total_tasks),
            istr(total_task_retries),
            istr(total_invocations),
        ]
        summary_str += (
            print_row(content, workflow_summary_col_size, output_format) + "\n"
        )

        content = [
            "Jobs",
            istr(total_succeeded_jobs),
            istr(total_failed_jobs),
            istr(total_unsubmitted_jobs),
            istr(total_jobs),
            str(total_job_retries),
            istr(total_job_instance_retries),
        ]
        summary_str += (
            print_row(content, workflow_summary_col_size, output_format) + "\n"
        )

        content = [
            "Sub-Workflows",
            istr(total_succeeded_sub_wfs),
            istr(total_failed_sub_wfs),
            istr(total_unsubmitted_sub_wfs),
            istr(total_sub_wfs),
            str(total_sub_wfs_retries),
            istr(total_sub_wfs_tries),
        ]
        summary_str += (
            print_row(content, workflow_summary_col_size, output_format) + "\n"
        )

        if output_format == "text":
            summary_str += "".center(sum(workflow_summary_col_size), "-") + "\n\n"

    if time_summary is True:
        states = workflow_stats.get_workflow_states()
        wwt = stats_utils.get_workflow_wall_time(states)
        wcjwt, wcgpt, wcbpt = workflow_stats.get_workflow_cum_job_wall_time()
        ssjwt, ssgpt, ssbpt = workflow_stats.get_submit_side_job_wall_time()
        if output_format == "text":

            def myfmt(val):
                if val is None:
                    return "-"
                else:
                    return format_seconds(val)

            if not multiple_wf:
                summary_str += "{:<57}: {}\n".format("Workflow wall time", myfmt(wwt))
            summary_str += "{:<57}: {}\n".format(
                "Cumulative job wall time", myfmt(wcjwt)
            )
            summary_str += "{:<57}: {}\n".format(
                "Cumulative job wall time as seen from submit side", myfmt(ssjwt),
            )
            summary_str += "{:<57}: {}\n".format(
                "Cumulative job badput wall time", myfmt(wcbpt),
            )
            summary_str += "{:<57}: {}\n".format(
                "Cumulative job badput wall time as seen from submit side",
                myfmt(ssbpt),
            )
        else:

            def myfmt(val):
                if val is None:
                    return ""
                else:
                    return str(val)

            summary_str += (
                print_row(workflow_time_summary_col_name_csv, None, output_format)
                + "\n"
            )
            summary_str += "workflow_wall_time,%s\n" % myfmt(wwt)
            summary_str += "workflow_cumulative_job_wall_time,%s\n" % myfmt(wcjwt)
            summary_str += "cumulative_job_walltime_from_submit_side,%s\n" % myfmt(
                ssjwt
            )
            summary_str += "workflow_cumulative_badput_time,%s\n" % myfmt(wcbpt)
            summary_str += (
                "cumulative_job_badput_walltime_from_submit_side,%s\n" % myfmt(ssbpt)
            )

        # PM-1260 integrity metrics summary
        int_metrics_summary = workflow_stats.get_summary_integrity_metrics()
        if int_metrics_summary and output_format == "text":

            def myfmt(val):
                if val is None:
                    return "-"
                else:
                    return format_seconds(val)

            summary_str += """
# Integrity Metrics
# Number of files for which checksums were compared/computed along with total time spent doing it. \n"""
            for result in int_metrics_summary:
                type = result.type
                if result.type == "check":
                    type = "compared"
                if result.type == "compute":
                    type = "generated"
                summary_str += "{} files checksums {} with total duration of {}\n".format(
                    result.count, type, myfmt(result.duration),
                )

            summary_str += """
# Integrity Errors
# Total:
#       Total number of integrity errors encountered across all job executions(including retries) of a workflow.
# Failures:
#       Number of failed jobs where the last job instance had integrity errors.
"""
            int_error_summary = workflow_stats.get_tag_metrics("int.error")
            for result in int_error_summary:
                type = result.name
                if result.name == "int.error":
                    type = "integrity"
                summary_str += (
                    "Total:    A total of %s %s errors encountered in the workflow\n"
                    % (result.count, type)
                )

            # PM-1295 jobs where last_job_instance failed and there were integrity errors in those last job instances
            summary_str += "Failures: %s job failures had integrity errors\n" % (
                total_failed_jobs_integrity.failed or 0
            )

    return summary_str


def print_individual_workflow_stats(
    workflow_stats, workflow_id, dax_label, output_format
):
    """
    Prints the workflow statistics of workflow
    @param workflow_stats :  workflow statistics object reference
    @param workflow_id  : workflow_id (title of the workflow table)
    """
    content_str = "\n"
    # individual workflow status

    # Add dax_label to workflow_id if writing text file
    if output_format == "text" and dax_label != "":
        workflow_id = workflow_id + " (" + dax_label + ")"

    # workflow status
    workflow_stats.set_job_filter("all")
    total_wf_retries = workflow_stats.get_workflow_retries()
    # only used for the text output...
    content = [workflow_id, istr(total_wf_retries)]
    retry_col_size = workflow_status_col_size[len(workflow_status_col_size) - 1]
    wf_status_str = print_row(
        content,
        [sum(workflow_status_col_size) - retry_col_size, retry_col_size],
        output_format,
    )

    # tasks
    workflow_stats.set_job_filter("nonsub")
    total_tasks = workflow_stats.get_total_tasks_status()
    total_succeeded_tasks = workflow_stats.get_total_succeeded_tasks_status(uses_PMC)
    total_failed_tasks = workflow_stats.get_total_failed_tasks_status()
    total_unsubmitted_tasks = total_tasks - (total_succeeded_tasks + total_failed_tasks)
    total_task_retries = workflow_stats.get_total_tasks_retries()
    total_task_invocations = (
        total_succeeded_tasks + total_failed_tasks + total_task_retries
    )
    if output_format == "text":
        content = [
            "Tasks",
            istr(total_succeeded_tasks),
            istr(total_failed_tasks),
            istr(total_unsubmitted_tasks),
            istr(total_tasks),
            istr(total_task_retries),
            istr(total_task_invocations),
            "",
        ]
    else:
        content = [
            workflow_id,
            dax_label,
            "Tasks",
            istr(total_succeeded_tasks),
            istr(total_failed_tasks),
            istr(total_unsubmitted_tasks),
            istr(total_tasks),
            istr(total_task_retries),
            istr(total_task_invocations),
            istr(total_wf_retries),
        ]

    tasks_status_str = print_row(content, workflow_status_col_size, output_format)

    # job status
    workflow_stats.set_job_filter("nonsub")
    total_jobs = workflow_stats.get_total_jobs_status()

    tmp = workflow_stats.get_total_succeeded_failed_jobs_status()
    total_succeeded_jobs = tmp.succeeded
    total_failed_jobs = tmp.failed

    total_unsubmitted_jobs = total_jobs - (total_succeeded_jobs + total_failed_jobs)
    total_job_retries = workflow_stats.get_total_jobs_retries()
    total_job_invocations = total_succeeded_jobs + total_failed_jobs + total_job_retries
    if output_format == "text":
        content = [
            "Jobs",
            istr(total_succeeded_jobs),
            istr(total_failed_jobs),
            istr(total_unsubmitted_jobs),
            istr(total_jobs),
            istr(total_job_retries),
            istr(total_job_invocations),
            "",
        ]
    else:
        content = [
            workflow_id,
            dax_label,
            "Jobs",
            istr(total_succeeded_jobs),
            istr(total_failed_jobs),
            istr(total_unsubmitted_jobs),
            istr(total_jobs),
            istr(total_job_retries),
            istr(total_job_invocations),
            istr(total_wf_retries),
        ]

    jobs_status_str = print_row(content, workflow_status_col_size, output_format)

    # sub workflow
    workflow_stats.set_job_filter("subwf")
    total_sub_wfs = workflow_stats.get_total_jobs_status()

    tmp = workflow_stats.get_total_succeeded_failed_jobs_status()

    total_succeeded_sub_wfs = 0
    if tmp.succeeded:
        total_succeeded_sub_wfs = tmp.succeeded

    total_failed_sub_wfs = 0
    if tmp.failed:
        total_failed_sub_wfs = tmp.failed

    total_unsubmitted_sub_wfs = total_sub_wfs - (
        total_succeeded_sub_wfs + total_failed_sub_wfs
    )
    total_sub_wfs_retries = workflow_stats.get_total_jobs_retries()
    total_sub_wfs_invocations = (
        total_succeeded_sub_wfs + total_failed_sub_wfs + total_sub_wfs_retries
    )
    if output_format == "text":
        content = [
            "Sub Workflows",
            istr(total_succeeded_sub_wfs),
            istr(total_failed_sub_wfs),
            istr(total_unsubmitted_sub_wfs),
            istr(total_sub_wfs),
            istr(total_sub_wfs_retries),
            istr(total_sub_wfs_invocations),
            "",
        ]
    else:
        content = [
            workflow_id,
            dax_label,
            "Sub_Workflows",
            istr(total_succeeded_sub_wfs),
            istr(total_failed_sub_wfs),
            istr(total_unsubmitted_sub_wfs),
            istr(total_sub_wfs),
            istr(total_sub_wfs_retries),
            istr(total_sub_wfs_invocations),
            istr(total_wf_retries),
        ]

    sub_wf_status_str = print_row(content, workflow_status_col_size, output_format)

    if output_format == "text":
        # Only print these in the text format output
        content_str += "".center(sum(workflow_status_col_size), "-") + "\n"
        content_str += wf_status_str + "\n"

    content_str += tasks_status_str + "\n"
    content_str += jobs_status_str + "\n"
    content_str += sub_wf_status_str + "\n"

    return content_str


def print_individual_wf_job_stats(
    workflow_stats, workflow_id, dax_label, output_format
):
    """
    Prints the job statistics of workflow
    @param workflow_stats :  workflow statistics object reference
    @param workflow_id : workflow_id (title for the table)
    """
    job_stats_list = []
    job_retry_count_dict = {}

    # Add dax_label to workflow_id if writing text file
    if output_format == "text":
        workflow_id = workflow_id + " (" + dax_label + ")"

    if output_format == "text":
        job_status_str = "\n# " + workflow_id + "\n"
    else:
        job_status_str = "\n"

    if output_format == "text":
        max_length = [max(0, len(i)) for i in job_stats_col_name_text]

    wf_job_stats_list = workflow_stats.get_job_statistics()

    # Go through each job in the workflow
    for job in wf_job_stats_list:
        job_stats = JobStatistics()
        job_stats.name = job.job_name
        job_stats.site = job.site
        job_stats.kickstart = job.kickstart
        job_stats.multiplier_factor = job.multiplier_factor
        job_stats.kickstart_mult = job.kickstart_multi
        job_stats.remote_cpu_time = job.remote_cpu_time
        job_stats.post = job.post_time
        job_stats.runtime = job.runtime
        job_stats.condor_delay = job.condor_q_time
        job_stats.resource = job.resource_delay
        job_stats.seqexec = job.seqexec
        job_stats.exitcode = utils.raw_to_regular(job.exit_code)
        job_stats.hostname = job.host_name
        if job_stats.seqexec is not None and job_stats.kickstart is not None:
            job_stats.seqexec_delay = float(job_stats.seqexec) - float(
                job_stats.kickstart
            )
        if job.job_name in job_retry_count_dict:
            job_retry_count_dict[job.job_name] += 1
        else:
            job_retry_count_dict[job.job_name] = 1
        job_stats.retry_count = job_retry_count_dict[job.job_name]

        if output_format == "text":
            max_length[0] = max(max_length[0], len(job_stats.name))
            max_length[1] = max(max_length[1], len(str(job_stats.retry_count)))
            max_length[2] = max(max_length[2], len(job_stats.site or " "))
            max_length[3] = max(max_length[3], len(str(job_stats.kickstart)))
            max_length[4] = max(max_length[4], len(str(job_stats.multiplier_factor)))
            max_length[5] = max(max_length[5], len(str(job_stats.kickstart_mult)))
            max_length[6] = max(max_length[6], len(str(job_stats.remote_cpu_time)))
            max_length[7] = max(max_length[7], len(str(job_stats.post)))
            max_length[8] = max(max_length[8], len(str(job_stats.condor_delay)))
            max_length[9] = max(max_length[9], len(str(job_stats.resource)))
            max_length[10] = max(max_length[10], len(str(job_stats.runtime)))
            max_length[11] = max(max_length[11], len(str(job_stats.seqexec)))
            max_length[12] = max(max_length[12], len(str(job_stats.seqexec_delay)))
            max_length[13] = max(max_length[13], len(str(job_stats.exitcode)))
            max_length[14] = max(
                max_length[14],
                len(job_stats.hostname if job_stats.hostname else "None"),
            )

        job_stats_list.append(job_stats)

    # Print header
    if output_format == "text":
        max_length = [i + 1 for i in max_length]
        job_status_str += print_row(job_stats_col_name_text, max_length, output_format)
    else:
        job_status_str += print_row(
            job_stats_col_name_csv, job_stats_col_size, output_format
        )

    job_status_str += "\n"

    # printing
    # find the pretty print length
    for job_stat in job_stats_list:
        job_det = job_stat.getFormattedJobStatistics()
        if output_format == "text":
            index = 0
            for content in job_det:
                job_status_str += str(content).ljust(max_length[index])
                index = index + 1
        else:
            job_status_str += workflow_id
            job_status_str += ","
            job_status_str += dax_label
            for content in job_det:
                job_status_str += "," + str(content)

        job_status_str += NEW_LINE_STR

    return job_status_str


def print_wf_transformation_stats(stats, workflow_id, dax_label, fmt):
    """
    Prints the transformation statistics of workflow
    stats       : workflow statistics object reference
    workflow_id : UUID of workflow
    dax_label   : Name of workflow
    format      : Format of report ('text' or 'csv')
    """
    if fmt not in ["text", "csv"]:
        print("Output format %s not recognized!" % fmt)
        sys.exit(1)

    report = ["\n"]

    if fmt == "text":
        # In text file, we need a line with the workflow id first
        report.append("# {} ({})".format(workflow_id, dax_label or "All"))

    col_names = transformation_stats_col_name_text
    if fmt == "csv":
        col_names = transformation_stats_col_name_csv

    transformation_statistics = stats.get_transformation_statistics()

    if fmt == "text":
        max_length = [max(0, len(col_names[i])) for i in range(13)]
        columns = ["" for i in range(13)]

        for t in transformation_statistics:
            max_length[0] = max(max_length[0], len(t.transformation))
            max_length[1] = max(max_length[1], len(str(t.type)))
            max_length[2] = max(max_length[2], len(str(t.count)))
            max_length[3] = max(max_length[3], len(str(t.min)))
            max_length[4] = max(max_length[4], len(str(t.max)))
            max_length[5] = max(max_length[5], len(str(t.avg)))
            max_length[6] = max(max_length[6], len(str(t.sum)))
            # maxrss
            max_length[7] = max(max_length[7], len(str(t.min_maxrss)))
            max_length[8] = max(max_length[8], len(str(t.max_maxrss)))
            max_length[9] = max(max_length[9], len(str(t.avg_maxrss)))
            # avg_cpu
            max_length[10] = max(max_length[10], len(str(t.min_avg_cpu)))
            max_length[11] = max(max_length[11], len(str(t.max_avg_cpu)))
            max_length[12] = max(max_length[12], len(str(t.avg_avg_cpu)))

        max_length = [i + 1 for i in max_length]

    header_printed = False

    for t in transformation_statistics:
        content = [
            t.transformation,
            t.type,
            str(t.count),
            fstr(t.min),
            fstr(t.max),
            fstr(t.avg),
            fstr(t.sum),
            fstr(t.min_maxrss / 1024) if t.min_maxrss else "-",
            fstr(t.max_maxrss / 1024) if t.max_maxrss else "-",
            fstr(t.avg_maxrss / 1024) if t.avg_maxrss else "-",
            pstr(t.min_avg_cpu * 100) if t.min_avg_cpu else "-",
            pstr(t.max_avg_cpu * 100) if t.max_avg_cpu else "-",
            pstr(t.avg_avg_cpu * 100) if t.avg_avg_cpu else "-",
        ]

        if fmt == "text":
            for i in range(0, 13):
                columns[i] = col_names[i].ljust(max_length[i])
                content[i] = content[i].ljust(max_length[i])

        if fmt == "csv":
            columns = transformation_stats_col_name_csv
            content = [workflow_id, dax_label] + content

        if not header_printed:
            header_printed = True
            report.append(print_row(columns, transformation_stats_col_size, fmt))

        report.append(print_row(content, transformation_stats_col_size, fmt))

    return NEW_LINE_STR.join(report) + NEW_LINE_STR


def print_wf_integrity_stats(stats, workflow_id, dax_label, fmt):
    """
    Prints the integrity statistics of workflow
    stats       : workflow statistics object reference
    workflow_id : UUID of workflow
    dax_label   : Name of workflow
    format      : Format of report ('text' or 'csv')
    """
    if fmt not in ["text", "csv"]:
        print("Output format %s not recognized!" % fmt)
        sys.exit(1)

    report = ["\n"]

    if fmt == "text":
        # In text file, we need a line with the workflow id first
        report.append("# {} ({})".format(workflow_id, dax_label or "All"))

    col_names = integrity_stats_col_name_text
    if fmt == "csv":
        col_names = integrity_stats_col_name_csv

    integrity_statistics = stats.get_integrity_metrics()

    if fmt == "text":
        max_length = [max(0, len(col_names[i])) for i in range(4)]
        columns = ["" for i in range(4)]
        # figure out max lengths?
        for i in integrity_statistics:
            max_length[0] = max(max_length[0], len(i.type))
            max_length[1] = max(max_length[0], len(i.file_type))
            max_length[2] = max(max_length[1], len(str(i.count)))
            max_length[3] = max(max_length[2], len(str(i.duration)))

        max_length = [i + 1 for i in max_length]

    header_printed = False

    for i in integrity_statistics:
        content = [i.type, i.file_type, str(i.count), str(i.duration)]

        if fmt == "text":
            for i in range(0, 4):
                columns[i] = col_names[i].ljust(max_length[i])
                content[i] = content[i].ljust(max_length[i])

        if fmt == "csv":
            columns = integrity_stats_col_name_csv
            content = [workflow_id, dax_label] + content

        if not header_printed:
            header_printed = True
            report.append(print_row(columns, integrity_stats_col_size, fmt))

        report.append(print_row(content, integrity_stats_col_size, fmt))

    return NEW_LINE_STR.join(report) + NEW_LINE_STR


def print_statistics_by_time_and_host(stats, fmt, combined=True, per_host=True):
    """
    Prints the job instance and invocation statistics sorted by time
    @param stats     : workflow statistics object reference
    @param fmt       : indicates how to format the output: "text" or "csv"
    @param combined  : print combined output (all hosts consolidated)
    @param per_host  : print per-host totals
    """
    report = []
    stats.set_job_filter("nonsub")
    stats.set_time_filter("hour")
    stats.set_transformation_filter(exclude=["condor::dagman"])

    if combined is True:
        col_names = time_stats_col_name_text
        if fmt == "csv":
            col_names = time_stats_col_name_csv

        report.append("\n# Job instances statistics per " + time_filter)
        report.append(print_row(col_names, time_stats_col_size, fmt))
        stats_by_time = stats.get_jobs_run_by_time()
        formatted = stats_utils.convert_stats_to_base_time(stats_by_time, time_filter)
        for s in formatted:
            content = [s["date_format"], str(s["count"]), fstr(s["runtime"])]
            if fmt == "csv":
                content.insert(0, "jobs/" + time_filter)
            report.append(print_row(content, time_stats_col_size, fmt))

        report.append("\n# Invocation statistics run per " + time_filter)
        report.append(print_row(col_names, time_stats_col_size, fmt))
        stats_by_time = stats.get_invocation_by_time()
        formatted = stats_utils.convert_stats_to_base_time(stats_by_time, time_filter)
        for s in formatted:
            content = [s["date_format"], str(s["count"]), fstr(s["runtime"])]
            if fmt == "csv":
                content.insert(0, "invocations/" + time_filter)
            report.append(print_row(content, time_stats_col_size, fmt))

    if per_host is True:
        col_names = time_host_stats_col_name_text
        if fmt == "csv":
            col_names = time_host_stats_col_name_csv

        report.append("\n# Job instances statistics on host per " + time_filter)
        report.append(print_row(col_names, time_host_stats_col_size, fmt))
        stats_by_time = stats.get_jobs_run_by_time_per_host()
        formatted_stats_list = stats_utils.convert_stats_to_base_time(
            stats_by_time, time_filter, True
        )
        for s in formatted_stats_list:
            content = [
                s["date_format"],
                str(s["host"]),
                str(s["count"]),
                fstr(s["runtime"]),
            ]
            if fmt == "csv":
                content.insert(0, "jobs/host/" + time_filter)
            report.append(print_row(content, time_host_stats_col_size, fmt))

        report.append("\n# Invocation statistics on host per " + time_filter)
        report.append(print_row(col_names, time_host_stats_col_size, fmt))
        stats_by_time = stats.get_invocation_by_time_per_host()
        formatted_stats_list = stats_utils.convert_stats_to_base_time(
            stats_by_time, time_filter, True
        )
        for s in formatted_stats_list:
            content = [
                s["date_format"],
                str(s["host"]),
                str(s["count"]),
                fstr(s["runtime"]),
            ]
            if fmt == "csv":
                content.insert(0, "invocations/host/" + time_filter)
            report.append(print_row(content, time_host_stats_col_size, fmt))

    return "\n".join(report)


def main():
    # Configure command line option parser
    prog_usage = (
        "%s [options] [[SUBMIT_DIRECTORY ..] | [WORKFLOW_UUID ..]]" % sys.argv[0]
    )

    parser = optparse.OptionParser(usage=prog_usage)
    parser.add_option(
        "-o",
        "--output",
        action="store",
        dest="output_dir",
        help="Writes the output to given directory.",
    )
    parser.add_option(
        "-f",
        "--file",
        action="store",
        dest="filetype",
        choices=[FILE_TYPE_TXT, FILE_TYPE_CSV],
        default=FILE_TYPE_TXT,
        help="Select output file type. Valid values are 'text' and 'csv'. Default is '%default'.",
    )
    parser.add_option(
        "-c",
        "--conf",
        action="store",
        type="string",
        dest="config_properties",
        default=None,
        help="Specifies the properties file to use. This option overrides all other property files.",
    )
    parser.add_option(
        "-s",
        "--statistics-level",
        action="store",
        dest="statistics_level",
        # choices=['all', 'summary', 'wf_stats', 'jb_stats', 'tf_stats', 'ti_stats','int_stats'],
        default="summary",
        help="Comma separated list. Valid levels are: all,summary,wf_stats,jb_stats,tf_stats,ti_stats,int_stats; Default is '%default'.",
    )
    parser.add_option(
        "-t",
        "--time-filter",
        action="store",
        dest="time_filter",
        choices=["day", "hour"],
        default="day",
        help="Valid levels are: day,hour; Default is '%default'.",
    )
    parser.add_option(
        "-i",
        "--ignore-db-inconsistency",
        action="store_true",
        default=False,
        dest="ignore_db_inconsistency",
        help="turn off the check for db consistency",
    )
    parser.add_option(
        "-v",
        "--verbose",
        action="count",
        default=0,
        dest="verbose",
        help="Increase verbosity, repeatable",
    )
    parser.add_option(
        "-q",
        "--quiet",
        action="count",
        default=0,
        dest="quiet",
        help="Decrease verbosity, repeatable",
    )

    parser.add_option(
        "-m",
        "--multiple-wf",
        action="store_true",
        dest="multiple_wf",
        default=False,
        help="Calculate statistics for multiple workflows",
    )
    parser.add_option(
        "-p",
        "--ispmc",
        action="store_true",
        dest="is_pmc",
        default=False,
        help="Calculate statistics for workflows which use PMC",
    )
    parser.add_option(
        "-u",
        "--isuuid",
        action="store_true",
        dest="is_uuid",
        default=False,
        help="Set if the positional arguments are wf uuids",
    )

    # Parse command line options
    (options, args) = parser.parse_args()

    options.statistics_level = {
        t.lower().strip() for t in options.statistics_level.split(",")
    }

    if not options.statistics_level:
        options.statistics_level = {"summary"}

    sl = options.statistics_level - {
        "all",
        "summary",
        "wf_stats",
        "jb_stats",
        "tf_stats",
        "ti_stats",
        "int_stats",
    }
    if sl:
        sys.stderr.write(
            "Invalid value(s) for statistics_level, ignoring %s\n" % ",".join(sl)
        )

    # Multiple workflow is set to true if there are multiple positional arguments.
    multiple_wf = options.multiple_wf

    if len(args) < 1:
        # * means all workflows in the database, and . means current directory
        submit_dir = "."
        if multiple_wf:
            submit_dir = "*"
    elif len(args) > 1:
        options.multiple_wf = True
        multiple_wf = True
        submit_dir = args
    else:
        options.multiple_wf = False
        multiple_wf = False
        submit_dir = args[0]

    log_level = options.verbose - options.quiet
    if log_level < 0:
        root_logger.setLevel(logging.ERROR)
    elif log_level == 0:
        root_logger.setLevel(logging.WARNING)
    elif log_level == 1:
        root_logger.setLevel(logging.INFO)
    elif log_level > 1:
        root_logger.setLevel(logging.DEBUG)

    def check_dump(dir):
        if not os.path.isfile(
            os.path.join(dir, "braindump.yml")
        ) and not os.path.isfile(os.path.join(dir, "braindump.txt")):
            sys.stderr.write("Not a workflow submit directory: %s\n" % submit_dir)
            sys.exit(1)

    if multiple_wf:
        # Check for braindump file's existence if workflows are not specified as UUIDs and
        # statistics need to be calculated only on a sub set of workflows
        if not options.is_uuid and submit_dir != "*":
            for dir in submit_dir:
                check_dump(dir)
    else:
        if not options.is_uuid:
            check_dump(submit_dir)

    if options.ignore_db_inconsistency:
        logger.warning("Ignoring db inconsistency")
        logger.warning(
            "The tool is meant to be run after the completion of workflow run."
        )
    else:

        def loading_complete(dir):
            if not utils.loading_completed(dir):
                if utils.monitoring_running(dir):
                    sys.stderr.write(
                        "pegasus-monitord still running. Please wait for it to complete.\n"
                    )
                else:
                    sys.stderr.write("Please run pegasus monitord in replay mode.\n")
                sys.exit(1)

        if multiple_wf:
            if submit_dir == "*":
                logger.warning(
                    "Statistics have to be calculated on all workflows. Tool cannot check to see if all of them have finished. Ensure that all workflows have finished"
                )

            if not options.is_uuid and submit_dir != "*":
                for dir in submit_dir:
                    loading_complete(dir)
        else:
            if not options.is_uuid:
                loading_complete(submit_dir)

    # Figure out what statistics we need to calculate
    global calc_wf_stats
    global calc_wf_summary
    global calc_jb_stats
    global calc_tf_stats
    global calc_ti_stats
    global calc_int_stats
    sl = options.statistics_level
    logger.info("Statistics level is %s" % sl)
    if "all" in sl:
        calc_wf_stats = True
        calc_wf_summary = True
        calc_tf_stats = True
        calc_int_stats = True
        calc_ti_stats = True
        if not multiple_wf:
            calc_jb_stats = True

    if "summary" in sl:
        calc_wf_summary = True

    if "wf_stats" in sl:
        calc_wf_stats = True

    if "jb_stats" in sl:
        if multiple_wf:
            logger.fatal(
                "Job breakdown statistics cannot be computed over multiple workflows"
            )
            sys.exit(1)
        calc_jb_stats = True

    if "tf_stats" in sl:
        calc_tf_stats = True

    if "ti_stats" in sl:
        calc_ti_stats = True

    if "int_stats" in sl:
        calc_int_stats = True

    global file_type
    file_type = options.filetype
    logger.info("File type is %s" % file_type)

    global time_filter
    time_filter = options.time_filter
    logger.info("Time filter is %s" % time_filter)

    # Change the legend to show the time filter format
    tf_format = str(stats_utils.get_date_print_format(time_filter))

    time_stats_col_name_text[0] += tf_format
    time_stats_col_name_csv[1] += tf_format
    time_host_stats_col_name_text[0] += tf_format
    time_host_stats_col_name_csv[1] += tf_format

    if options.output_dir:
        delete_if_exists = False
        output_dir = options.output_dir
    else:
        delete_if_exists = True
        if multiple_wf or options.is_uuid:
            sys.stderr.write(
                "Output directory option is required when calculating statistics over multiple workflows.\n"
            )
            sys.exit(1)
        else:
            output_dir = os.path.join(submit_dir, DEFAULT_OUTPUT_DIR)

    logger.info("Output directory is %s" % output_dir)
    utils.create_directory(output_dir, delete_if_exists=delete_if_exists)

    global uses_PMC

    def use_pmc(dir):
        braindb = utils.slurp_braindb(dir)
        return braindb["uses_pmc"] is True

    if options.is_pmc:
        logger.info("Calculating statistics with use of PMC clustering")
        uses_PMC = True
    else:
        if options.is_uuid:
            # User provided workflow UUID
            logger.info("Workflows are specified as UUIDs and ispmc option is not set.")
            uses_PMC = False
        else:
            # User provided workflow submit directories
            if multiple_wf:
                if submit_dir == "*":
                    logger.info(
                        "Calculating statistics over all workflows, and ispmc option is not set."
                    )
                else:
                    # int(True) -> 1
                    tmp = sum(int(use_pmc(dir)) for dir in submit_dir)

                    # All workflow are either PMC or non PMC workflows?
                    if tmp == len(submit_dir) or tmp == 0:
                        uses_PMC = use_pmc(submit_dir[0])
                    else:
                        uses_PMC = False
                        logger.warn(
                            "Input workflows use both PMC & regular clustering! Calculating statistics with regular clustering"
                        )

            else:
                uses_PMC = use_pmc(submit_dir)

    # Check db_url, and get wf_uuid's
    if multiple_wf:
        if options.is_uuid or submit_dir == "*":
            # URL picked from config_properties file.
            output_db_url = connection.url_by_properties(
                options.config_properties, connection.DBType.WORKFLOW
            )
            wf_uuid = submit_dir

            if not output_db_url:
                logger.error(
                    'Unable to determine database URL. Kindly specify a value for "pegasus.monitord.output" property'
                )
                sys.exit(1)
        else:
            db_url_set = set()
            wf_uuid = []

            for dir in submit_dir:
                db_url = connection.url_by_submitdir(
                    dir, connection.DBType.WORKFLOW, options.config_properties
                )
                uuid = connection.get_wf_uuid(dir)
                db_url_set.add(db_url)
                wf_uuid.append(uuid)

            if len(db_url_set) != 1:
                logger.error(
                    "Workflows are distributed across multiple databases, which is not supported"
                )
                sys.exit(1)

            output_db_url = db_url_set.pop()

    else:
        if options.is_uuid:
            output_db_url = connection.url_by_properties(
                options.config_properties, connection.DBType.WORKFLOW
            )
            wf_uuid = submit_dir

            if not output_db_url:
                logger.error(
                    'Unable to determine database URL. Kindly specify a value for "pegasus.monitord.output" property'
                )
                sys.exit(1)
        else:
            output_db_url = connection.url_by_submitdir(
                submit_dir, connection.DBType.WORKFLOW, options.config_properties
            )
            wf_uuid = connection.get_wf_uuid(submit_dir)

    logger.info("DB URL is: %s" % output_db_url)
    logger.info("workflow UUID is: %s" % wf_uuid)

    if output_db_url is not None:
        errors = print_workflow_details(
            output_db_url, wf_uuid, output_dir, multiple_wf=multiple_wf
        )

        if errors:
            logger.error("Failed to generate %d type(s) of statistics" % errors)
            sys.exit(1)


if __name__ == "__main__":
    main()



##
#  Copyright 2007-2012 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##
"""
This file implements the metadata related classes for pegasus-monitord.
"""


# Import Python modules
import json
import logging
import os
import tempfile
from io import StringIO

__author__ = "Karan Vahi <vahi@isi.edu>"

logger = logging.getLogger(__name__)


class Metadata:
    """
    Base class for for different types of metadata
    """

    def __init__(self):
        self._id = None
        self._type = None
        self._attributes = {}  # dictionary indexed by attribute name and values

    def get_id(self):
        return self._id

    def set_id(self, id):
        self._id = id

    def add_attribute(self, key, value):
        self._attributes[key] = value

    def get_type(self):
        return self._type

    def get_attribute_value(self, key):
        return self._attributes.get(key)

    def get_attribute_keys(self):
        return self._attributes.keys()

    @staticmethod
    def write_to_jsonfile(metadata_list, directory, name, prefix="pegasus-monitord"):
        try:
            temp_file = tempfile.NamedTemporaryFile(
                dir=directory, prefix=prefix, suffix=".meta", delete=False, mode="w"
            )
            jsonify(metadata_list, temp_file)
            logger.debug("Written out metadata to %s", temp_file.name)
            os.chmod(temp_file.name, 0o644)
            # rename the file to the name to assure atomicity
            os.rename(temp_file.name, os.path.join(directory, name))

        except Exception as e:
            # Error sending this event... disable the sink from now on...
            logger.warning(
                " unable to write out  metadata to directory %s with name %s ",
                directory,
                name,
            )
            logger.exception(e)
            return False

        return True


class FileMetadata(Metadata):
    """
    Represents a single mkdir request
    """

    def __init__(self):
        super().__init__()
        self._type = "file"

    def convert_to_rce(self):
        """
        Converts to a file based RC description
        :return:
        """

        rce = StringIO()
        rce.write(self.get_id())
        pfn = self.get_attribute_value("pfn")
        if pfn is None:
            pfn = "@@PFN@@"
        if pfn:
            rce.write(" ")
            rce.write(pfn)

        for key in self.get_attribute_keys():
            rce.write(" ")
            rce.write(key)
            rce.write("=")
            rce.write('"')
            rce.write(self.get_attribute_value(key))
            rce.write('"')
        return rce.getvalue()


class MetadataCustomEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Metadata):
            return o.__dict__
        return json.JSONEncoder.default(self, o)


def jsonify(*args, **kwargs):
    json.dump(*args, cls=MetadataCustomEncoder, indent=2, **kwargs)


def main():
    a = FileMetadata()
    a._id = "f.a"
    a.add_attribute("size", "100")
    a.add_attribute("checksum", "XXXXXX")

    b = FileMetadata()
    b._id = "f.b"
    b.add_attribute("size", "10")
    b.add_attribute("checksum", "XXXXXX")

    print(json.dumps(a.__dict__))

    l = []
    l.append(a)
    l.append(b)
    l.append(1)

    print(json.dumps(l, cls=MetadataCustomEncoder, indent=2))

    Metadata.write_to_jsonfile(l, "/tmp", "test")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
#  Copyright 2017-2021 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__author__ = "Gideon Juve"
__author__ = "Rafael Ferreira da Silva"

DB_VERSION = 3

import logging

from sqlalchemy.exc import *

from Pegasus.db.admin.admin_loader import *
from Pegasus.db.admin.versions.base_version import *

log = logging.getLogger(__name__)


class Version(BaseVersion):
    def __init__(self, connection):
        super().__init__(connection)

    def update(self, force=False):
        "Add plan_command field to ensemble_workflow table"
        log.info("Updating to version %s" % DB_VERSION)
        try:
            self.db.execute(
                "ALTER TABLE ensemble_workflow ADD plan_command VARCHAR(1024) NOT NULL default './plan.sh'"
            )
        except (OperationalError, ProgrammingError):
            pass
        except Exception as e:
            self.db.rollback()
            log.exception(e)
            raise Exception(e)

    def downgrade(self, force=False):
        "Downgrade is not necessary as plan_command is added with a default that works for old versions"



"""
Functions for output pegasus-monitord events to various destinations.
"""

##
#  Copyright 2007-2011 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##

import logging
import queue
import re
import socket
import ssl
import time
import traceback
import urllib.parse
from threading import Thread

from Pegasus import json
from Pegasus.db import connection, expunge
from Pegasus.db.dashboard_loader import DashboardLoader
from Pegasus.db.workflow_loader import WorkflowLoader
from Pegasus.netlogger import nlapi
from Pegasus.tools import properties, utils

log = logging.getLogger(__name__)

# Optional imports, only generate 'warnings' if they fail
bson = None
try:
    import bson
except Exception:
    log.info("cannot import BSON library, 'bson'")

amqp = None
try:
    import pika as amqp
except Exception:
    log.info("cannot import AMQP library")

# Event name-spaces
STAMPEDE_NS = "stampede."
DASHBOARD_NS = "dashboard."


def purge_wf_uuid_from_database(rundir, output_db):
    """
    This function purges a workflow id from the output database.
    """
    # PM-652 do nothing for sqlite
    # DB is already rotated in pegasus-monitord
    if output_db.lower().startswith("sqlite"):
        return

    # Parse the braindump file
    wfparams = utils.slurp_braindb(rundir)

    wf_uuid = wfparams.get("wf_uuid", None)
    if wf_uuid is None:
        return

    expunge.delete_workflow(output_db, wf_uuid)


def purge_wf_uuid_from_dashboard_database(rundir, output_db):
    """
    This function purges a workflow id from the output database.
    """

    # Parse the braindump file
    wfparams = utils.slurp_braindb(rundir)

    wf_uuid = wfparams.get("wf_uuid", None)
    if wf_uuid is None:
        return

    expunge.delete_dashboard_workflow(output_db, wf_uuid)


class OutputURL:
    """
    Break output URL into named parts for easier handling.
    """

    def __init__(self, url):

        (
            self.scheme,
            self.netloc,
            self.path,
            self.params,
            query,
            frag,
        ) = urllib.parse.urlparse(url)

        host_port = ""
        user_pass = ""

        if "@" in self.netloc:
            user_pass, host_port = self.netloc.split("@", 1)
        else:
            host_port = self.netloc
        if ":" in host_port:
            self.host, portstr = host_port.split(":", 1)
            self.port = int(portstr)
        else:
            self.host = self.netloc
            self.port = None

        if ":" in user_pass:
            self.user, self.password = user_pass.split(":", 1)


class EventSink:
    """
    Base class for an Event Sink.
    """

    def __init__(self):
        self._log = logging.getLogger(
            "{}.{}".format(self.__module__, self.__class__.__name__)
        )
        # Set listing events handled to be kept consistent with dict in workflow loader
        self._acceptedEvents = (
            "stampede.wf.plan",
            "stampede.wf.map.task_job",
            "stampede.static.start",
            "stampede.static.end",
            "stampede.xwf.start",
            "stampede.xwf.end",
            "stampede.xwf.map.subwf_job",
            "stampede.task.info",
            "stampede.task.edge",
            "stampede.job.info",
            "stampede.job.edge",
            "stampede.job_inst.pre.start",
            "stampede.job_inst.pre.term",
            "stampede.job_inst.pre.end",
            "stampede.job_inst.submit.start",
            "stampede.job_inst.submit.end",
            "stampede.job_inst.held.start",
            "stampede.job_inst.held.end",
            "stampede.job_inst.main.start",
            "stampede.job_inst.main.term",
            "stampede.job_inst.main.end",
            "stampede.job_inst.post.start",
            "stampede.job_inst.post.term",
            "stampede.job_inst.post.end",
            "stampede.job_inst.host.info",
            "stampede.job_inst.image.info",
            "stampede.job_inst.abort.info",
            "stampede.job_inst.grid.submit.start",
            "stampede.job_inst.grid.submit.end",
            "stampede.job_inst.globus.submit.start",
            "stampede.job_inst.globus.submit.end",
            "stampede.job_inst.tag",
            "stampede.job_inst.composite",
            "stampede.inv.start",
            "stampede.inv.end",
            "stampede.static.meta.start",
            "stampede.xwf.meta",
            "stampede.task.meta",
            "stampede.rc.meta",
            "stampede.int.metric",
            "stampede.rc.pfn",
            "stampede.wf.map.file",
            "stampede.static.meta.end",
            "stampede.task.monitoring",
        )

    def send(self, event, kw):
        """
        Clients call this function to send an event to the sink.
        """

    def close(self):
        """
        Clients call this function to close the output to this sink.
        """

    def flush(self):
        "Clients call this to flush events to the sink"


class DBEventSink(EventSink):
    """
    Write wflow event logs to database via loader
    """

    def __init__(
        self,
        dest,
        db_stats=False,
        namespace=STAMPEDE_NS,
        props=None,
        db_type=None,
        backup=False,
        **kw
    ):
        self._namespace = namespace
        # pick the right database loader based on prefix
        if namespace == STAMPEDE_NS:
            self._db = WorkflowLoader(
                dest,
                perf=db_stats,
                batch=True,
                props=props,
                db_type=db_type,
                backup=backup,
            )
        elif namespace == DASHBOARD_NS:
            self._db = DashboardLoader(
                dest,
                perf=db_stats,
                batch=True,
                props=props,
                db_type=db_type,
                backup=backup,
            )
        else:
            raise ValueError("Unknown namespace specified '%s'" % (namespace))

        super().__init__()

    def send(self, event, kw):
        self._log.trace("send.start event=%s", event)
        d = {"event": self._namespace + event}
        for k, v in kw.items():
            d[k.replace("__", ".")] = v
        self._db.process(d)
        self._log.trace("send.end event=%s", event)

    def close(self):
        self._log.trace("close.start")
        self._db.finish()
        self._log.trace("close.end")

    def flush(self):
        self._db.flush()


class FileEventSink(EventSink):
    """
    Write wflow event logs to a file.
    """

    def __init__(self, path, restart=False, encoder=None, **kw):
        super().__init__()
        if restart:
            self._output = open(path, "w", 1)
        else:
            self._output = open(path, "a", 1)
        self._encoder = encoder

    def send(self, event, kw):
        self._log.trace("send.start event=%s", event)
        self._output.write(self._encoder(event=event, **kw))
        if self._encoder == json_encode:
            self._output.write("\n")
        self._log.trace("send.end event=%s", event)

    def close(self):
        self._log.trace("close.start")
        self._output.close()
        self._log.trace("close.end")


class TCPEventSink(EventSink):
    """
    Write wflow event logs to a host:port.
    """

    def __init__(self, host, port, encoder=None, **kw):
        super().__init__()
        self._encoder = encoder
        self._sock = socket.socket()
        self._sock.connect((host, port))

    def send(self, event, kw):
        self._log.trace("send.start event=%s", event)
        self._sock.send(self._encoder(event=event, **kw))
        self._log.trace("send.end event=%s", event)

    def close(self):
        self._log.trace("close.start")
        self._sock.close()
        self._log.trace("close.end")


class AMQPEventSink(EventSink):
    """
    Write wflow event logs to an AMQP server.
    """

    EXCH_OPTS = {"exchange_type": "topic", "durable": True, "auto_delete": False}
    DEFAULT_AMQP_VIRTUAL_HOST = "pegasus"  # should be /

    def __init__(
        self,
        host,
        port,
        exch=None,
        encoder=None,
        userid="guest",
        password="guest",
        virtual_host=DEFAULT_AMQP_VIRTUAL_HOST,
        ssl_enabled=False,
        props=None,
        connect_timeout=None,
        **kw
    ):
        super().__init__()
        self._log.info("Encoder used {} Properties received {}".format(encoder, props))
        self._encoder = encoder
        self._handled_events = set()
        self._handle_all_events = False
        self.configure_filters(props.property("events"))
        self._msg_queue = queue.Queue()
        self._stopping = False
        self._exch = exch
        self._conn = None

        if connect_timeout is None:
            # pick timeout from properties
            connect_timeout = props.property("timeout")
            if connect_timeout:
                connect_timeout = float(connect_timeout)

        # insecure ssl
        SSLOptions = None
        if ssl_enabled:
            context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            SSLOptions = amqp.SSLOptions(context)

        creds = amqp.PlainCredentials(userid, password)
        self._params = amqp.ConnectionParameters(
            host=host,
            port=port,
            ssl_options=SSLOptions,
            virtual_host=virtual_host,
            credentials=creds,
            blocked_connection_timeout=connect_timeout,
            heartbeat=None,
        )  # None -> negotiate heartbeat with the AMQP server

        # initialize worker thread in daemon and start it
        self._worker_thread = Thread(target=self.event_publisher, daemon=True)
        self._worker_thread.start()

    def event_publisher(self):
        full_event, event, data = (None, None, None)
        reconnect_attempts = 0
        while not self._stopping:
            try:
                self._log.info(
                    "Connecting to host: %s:%s virtual host: %s exchange: %s with user: %s ssl: %s"
                    % (
                        self._params.host,
                        self._params.port,
                        self._params.virtual_host,
                        self._exch,
                        self._params.credentials.username,
                        not self._params.ssl_options is None,
                    )
                )

                self._conn = amqp.BlockingConnection(self._params)
                self._channel = self._conn.channel()
                self._channel.exchange_declare(self._exch, **self.EXCH_OPTS)

                reconnect_attempts = 0

                while not self._stopping:
                    try:
                        # if variables are initialized we haven't sent them yet.
                        # don't retrieve a new event, send the old one
                        if (full_event is None) and (event is None) and (data is None):
                            full_event, event, data = self._msg_queue.get(timeout=5)
                        self._log.trace("send.start event=%s", full_event)
                        self._channel.basic_publish(
                            body=data, exchange=self._exch, routing_key=full_event
                        )
                        self._log.trace("send.end event=%s", event)
                        # reset vars
                        full_event, event, data = (None, None, None)
                        # mark item as processed
                        self._msg_queue.task_done()
                    except queue.Empty:
                        self._conn.process_data_events()  # keep up with the AMQP heartbeats
                        continue

            # Do not recover if connection was closed by broker
            except amqp.exceptions.ConnectionClosedByBroker as err:
                self._log.error(
                    "Connection to %s:%s was closed by Broker - Not Recovering"
                    % (self._params.host, self._params.port)
                )
                self._log.error("Broker closed connection with: %s, stopping..." % err)
                self._conn = None
                break
            # Do not recover on channel errors
            except amqp.exceptions.AMQPChannelError as err:
                self._log.error(
                    "Channel error at %s:%s - Not Recovering"
                    % (self._params.host, self._params.port)
                )
                self._log.error("Channel error: %s, stopping..." % err)
                self._conn = None
                break
            # Recover on all other connection errors if reconnect attempts is less than 5
            except amqp.exceptions.AMQPConnectionError:
                reconnect_attempts += 1
                if reconnect_attempts > 5:
                    self._log.info(
                        "Connection to %s:%s was closed - Not Recovering"
                        % (self._params.host, self._params.port)
                    )
                    break
                else:
                    self._log.info(
                        "Connection to %s:%s was closed - Will try to recover the connection"
                        % (self._params.host, self._params.port)
                    )
                    time.sleep((2 ** reconnect_attempts) * 10)
                    continue

        if not self._conn is None:
            self._log.trace("connection - close.start")
            self._conn.close()
            self._log.trace("connection - close.end")

    def configure_filters(self, events):
        event_regexes = set()

        if events is None:
            # add pre-configured specific events
            event_regexes.add(re.compile(STAMPEDE_NS + "job_inst.tag"))
            event_regexes.add(re.compile(STAMPEDE_NS + "job_inst.composite"))
            event_regexes.add(re.compile(STAMPEDE_NS + "inv.end"))
            event_regexes.add(re.compile(STAMPEDE_NS + "wf.plan"))
        else:
            for exp in events.split(","):
                if exp == "*":
                    # short circuit
                    self._handle_all_events = True
                    self._log.debug("Events Handled: All")
                    return
                else:
                    event_regexes.add(re.compile(exp))

        # go through each regex and match against accepted events once
        for regex in event_regexes:
            # go through each list of accepted events to check match
            for event in self._acceptedEvents:
                if regex.search(event) is not None:
                    self._handled_events.add(event)

        self._log.debug("Events Handled: %s", self._handled_events)

    def send(self, event, kw):
        if not self._worker_thread.is_alive():
            raise Exception("AMQP publisher thread is dead. Cannot send amqp events.")

        full_event = STAMPEDE_NS + event
        if self.ignore(full_event):
            return
        data = self._encoder(event=event, **kw)
        self._msg_queue.put((full_event, event, data))

    def ignore(self, event):
        if self._handle_all_events:
            # we want all events
            return False

        return event not in self._handled_events

    def close(self):
        if self._worker_thread.is_alive():
            self._log.trace("Waiting for queue to emtpy.")
            self._msg_queue.join()  # wait for queue to empty if worker is alive
        self._stopping = True
        self._log.trace("Waiting for publisher thread to exit.")
        self._worker_thread.join()
        self._log.trace("Publisher thread exited.")


class MultiplexEventSink(EventSink):
    """
    Sends events to multiple end points
    """

    def __init__(self, dest, enc, prefix=STAMPEDE_NS, props=None, **kw):
        super().__init__()
        self._endpoints = {}
        self._log.info("Multiplexed Event Sink Connection Properties  %s", props)
        for key in props.keyset():
            if key.endswith(".url"):
                sink_name = key[0 : key.rfind(".url")]

                # remove from our copy sink_name properties if they exist
                endpoint_props = properties.Properties(
                    props.propertyset(sink_name + ".", remove=True)
                )
                try:
                    self._endpoints[sink_name] = create_wf_event_sink(
                        props.property(key),
                        db_type=connection.DBType.WORKFLOW,
                        enc=enc,
                        prefix=prefix,
                        props=endpoint_props,
                        multiplexed=True,
                        **kw,
                    )

                except Exception:
                    self._log.error(
                        "[multiplex event sender] Unable to connect to endpoint %s with props %s . Disabling"
                        % (sink_name, endpoint_props)
                    )
                    self._log.error(traceback.format_exc())

    def send(self, event, kw):
        remove_endpoints = []
        for key in self._endpoints:
            sink = self._endpoints[key]
            try:
                sink.send(event, kw)
            except Exception:
                self._log.error(traceback.format_exc())
                self._log.error(
                    "[multiplex event sender] error sending event. Disabling endpoint %s"
                    % key
                )
                self.close_sink(sink)
                remove_endpoints.append(key)

        # remove endpoints that are disabled
        for key in remove_endpoints:
            del self._endpoints[key]

    def close(self):
        for key in self._endpoints:
            self._log.debug("[multiplex event sender] Closing endpoint %s" % key)
            self.close_sink(self._endpoints[key])

    def close_sink(self, sink):
        try:
            sink.close()
        except Exception:
            pass

    def flush(self):
        "Clients call this to flush events to the sink"
        for key in self._endpoints:
            self._log.debug("[multiplex event sender] Flushing endpoint %s" % key)
            self._endpoints[key].flush()


def bson_encode(event, **kw):
    """
    Adapt bson.dumps() to NetLogger's Log.write() signature.
    """
    kw["event"] = STAMPEDE_NS + event
    return bson.dumps(kw)


def json_encode(event, **kw):
    """
    Adapt bson.dumps() to NetLogger's Log.write() signature.
    """
    kw["event"] = STAMPEDE_NS + event

    # PM-1355 , PM-1365 replace all __ and . with _
    for k, v in list(kw.items()):
        new_key = k.replace(".", "_")
        new_key = new_key.replace("__", "_")
        kw[new_key] = kw.pop(k)

    return json.dumps(kw)


def create_wf_event_sink(
    dest, db_type, enc=None, prefix=STAMPEDE_NS, props=None, multiplexed=False, **kw
):
    """
    Create & return subclass of EventSink, chosen by value of 'dest'
    and parameterized by values (if any) in 'kw'.
    """

    if dest is None:
        return None

    # we only subset the properties and strip off prefix once
    if not multiplexed:
        sink_props = get_workflow_connect_props(props, db_type)
        # we delete from our copy pegasus.catalog.workflow.url as we want with default prefix
        if "url" in sink_props.keyset():
            del sink_props["url"]
            sink_props.property("default.url", dest)
    else:
        sink_props = props

    # PM-898 are additional URL's to populate specified
    if not multiplexed and multiplex(dest, prefix, props):
        sink_props.property("default.url", dest)
        # any properties that don't have a . , remap to default.propname
        for key in sink_props.keyset():
            if key.find(".") == -1:
                sink_props.property("default." + key, sink_props.property(key))
                del sink_props[key]

        return MultiplexEventSink(dest, enc, prefix, sink_props, **kw)

    url = OutputURL(dest)

    log.info("Connecting workflow event sink to %s" % dest)

    # Pick an encoder

    def pick_encfn(enc_name, namespace):
        if enc_name is None or enc_name == "bp":
            # NetLogger name=value encoding
            encfn = nlapi.Log(level=nlapi.Level.ALL, prefix=namespace)
        elif enc_name == "bson":
            # BSON
            if bson is None:
                raise Exception(
                    "BSON encoding selected, but cannot import bson library"
                )
            encfn = bson_encode
        elif enc_name == "json":
            encfn = json_encode
        else:
            raise ValueError("Unknown encoding '%s'" % (enc_name))
        return encfn

    # Branch on scheme
    if url.scheme == "":
        sink = FileEventSink(dest, encoder=pick_encfn(enc, prefix), **kw)
        _type, _name = "file", dest
    elif url.scheme == "file":
        sink = FileEventSink(url.path, encoder=pick_encfn(enc, prefix), **kw)
        _type, _name = "file", url.path
    elif url.scheme == "x-tcp":
        if url.port is None:
            url.port = 14380
        sink = TCPEventSink(url.host, url.port, encoder=pick_encfn(enc, prefix), **kw)
        _type, _name = "network", "{}:{}".format(url.host, url.port)
    elif url.scheme in ["amqp", "amqps"]:
        # amqp://[USERNAME:PASSWORD@]<hostname>[:port]/[<virtualhost>]/<exchange_name>
        if amqp is None:
            raise Exception("AMQP destination selected, but cannot import AMQP library")
        if url.port is None:
            if url.scheme == "amqps":
                url.port = 5671  # RabbitMQ default TLS
            else:
                url.port = 5672  # RabbitMQ default

        # PM-1258 parse exchange and virtual host info
        exchange = None
        virtual_host = None
        path_comp = url.path.split("/")
        if path_comp is not None:
            exchange = path_comp.pop()
        if path_comp is not None:
            virtual_host = path_comp.pop()
            if len(virtual_host) == 0:
                virtual_host = None

        # PM-1355 set encoder to json always for AMQP endpoints
        enc = "json"
        sink = AMQPEventSink(
            url.host,
            url.port,
            virtual_host=virtual_host,
            exch=exchange,
            userid=url.user,
            password=url.password,
            ssl_enabled=(url.scheme == "amqps"),
            encoder=pick_encfn(enc, prefix),
            props=sink_props,
            **kw,
        )
        _type, _name = "AMQP", "{}:{}/{}".format(url.host, url.port, url.path)
    else:
        # load the appropriate DBEvent on basis of prefix passed
        sink = DBEventSink(dest, namespace=prefix, props=sink_props, **kw)
        _type, _name = "DB", dest

    log.info("output type={} namespace={} name={}".format(_type, prefix, _name))

    return sink


def multiplex(dest, prefix, props=None):
    """
    Determines whether we need to multiplex and events to multiple sinks
    :param dest:
    :param props:
    :return:
    """
    if props is None:
        return False

    # we never attempt multiplex on dashboard sink
    if prefix == DASHBOARD_NS:
        return False

    additional_sink_props = props.propertyset("pegasus.catalog.workflow" + ".", False)
    multiplex = False
    for key in additional_sink_props:
        if key == "pegasus.catalog.workflow.url":
            pass
        if key.endswith(".url"):
            multiplex = True
            break
    return multiplex


def get_workflow_connect_props(props, db_type):
    """
    Returns connection properties for workflow database

    :param props:
    :return:
    """

    if props is None:
        return None

    # first get the default one's with the star notation
    connect_props = properties.Properties(props.propertyset("pegasus.catalog.*.", True))

    prefix = "pegasus.catalog.workflow"
    if db_type == connection.DBType.MASTER:
        prefix = "pegasus.catalog.dashboard"

    # over ride these with workflow specific or dashboard specific props
    addons = props.propertyset(prefix + ".", True)
    for key in addons:
        connect_props.property(key, addons[key])

    return connect_props



"""
Common code for NetLogger parsers
"""
__author__ = "Dan Gunter <dkgunter@lbl.gov>"
__rcsid__ = "$Id: base.py 28287 2011-08-18 03:42:53Z dang $"
import calendar
import re
import time
from select import select

#
from Pegasus.netlogger import nlapi, nldate
from Pegasus.netlogger.nlapi import EVENT_FIELD, HASH_FIELD, TS_FIELD, Level
from Pegasus.netlogger.nllog import DoesLogging
from Pegasus.netlogger.parsers import nlreadline
from Pegasus.netlogger.util import hash_event

#
try:
    from pyparsing import (
        CharsNotIn,
        Each,
        Group,
        Literal,
        OneOrMore,
        ParseException,
        QuotedString,
        StringEnd,
        White,
        Word,
        ZeroOrMore,
        alphanums,
        oneOf,
    )

    HAVE_PYPARSING = True
except ImportError:
    HAVE_PYPARSING = False

# Special result code for parsers to return
# when they 'on purpose' skip a line
LINE_SKIPPED = 999

try:
    from hashlib import md5

    md5_new = md5
except ImportError:
    import md5

    md5_new = md5.new


def getGuid(*strings):
    m5 = md5_new()
    for s in strings:
        m5.update(s)
    t = m5.hexdigest()
    guid = "{}-{}-{}-{}-{}".format(t[:8], t[8:12], t[12:16], t[16:20], t[20:])
    return guid


def autoParseValue(vstr):
    try:
        value = int(vstr)
    except ValueError:
        try:
            value = float(vstr)
        except ValueError:
            value = vstr
    return value


def parse_ts(ts):
    "Parse a netlogger timestamp"
    # until 2033, a '1' in the first place means a float
    if ts[0] == "1":
        return float(ts)
    ts, subs = ts.split(".")
    subs = float("." + subs[:-1])
    return calendar.timegm(time.strptime(ts, r"%Y-%m-%dT%H:%M:%S")) + subs


parseDate = parse_ts


def tolerateBlanksAndComments(line=None, error=None, linenum=0):
    """Callback function fitting the signature of the callback
    expected by NLBaseParser.parseStream() that re-raises the error unless
    the line is empty or starts with a hash character, in which
    case it does nothing.
    """
    if len(line) <= 1 or line[0] == "#":
        pass
    else:
        raise (error)


# Parse BP log lines with Regular Expressions


class BPError(ValueError):
    """
    Exception class to indicate violations from the logging best practices
    standard.
    """

    def __init__(self, lineno, msg):
        """Create error object.

        Arguments:
          lineno - The line number on which the error occured.
          msg    - The error message.
        """
        self.lineno = lineno
        self.msg = msg

    def __str__(self):
        return "Parser error on line %i: %s" % (self.lineno, self.msg)


class BPValidationError(BPError):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return "Validation error: " + str(self.msg)


class BPParseError(BPError):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return "Parse error: " + str(self.msg)


"""
Regular expression that captures names and data
"""
BP_EXPR = re.compile(
    r"""
(?:
    \s*                        # leading whitespace
    ([0-9a-zA-Z_.\-]+)         # Name
    =
    (?:                        # Value:
      ([^"\s]+) |              # a) simple value
      "((?:[^"] | (?<=\\)")*)" # b) quoted string
    )
    \s*
)
""",
    flags=re.X,
)

"""
For validation, regular expression that captures all valid characters,
so by simply comparing with string length we can tell if there are invalid
characters in the string.
"""
BP_EXPR_WS = re.compile(
    r"""
(?:
    (\s*)                      # leading whitespace
    ([0-9a-zA-Z_.\-]+)         # Name
    =
    (?:                        # Value:
      ([^"\s]+) |              # a) simple value
      (")((?:[^"] | (?<=\\)")*)(") # b) quoted string
    )
    (\s*)
)""",
    flags=re.X,
)


def _bp_extract(s, as_dict=True, validate=False):
    """Parse BP log line and extract key=value pairs.

    If validate is False, this will skip over many types of "junk" data.
    The only escape sequence recognized is a backslash-escaped double-quote,
    within a quoted value.

    Args:
      s - Input line
      as_dict - If True, return a dictionary, otherwise, a list of tuples.
                Note: when OrderedDict (Python 3.1+, 2.7+) becomes common,
                this will go away.
      validate - If True, see if there are any 'extra' characters in the string.
                 If there are, raise a BPValidationError.
    Raises:
      BPParseError - If a token doesn't contain a recognizable name=value pair,
      BPValidationError - If (optional) validation fails.
    Returns:
      Dictionary or list of tuples (see `as_dict` arg)
    """
    if as_dict:
        result = {}
    else:
        result = []
    if not validate:
        for n, v, vq in BP_EXPR.findall(s):
            # check input
            if not n:
                raise BPParseError("Bad key: '{}'".format(n))
            # add to result
            if vq:
                v = vq.replace('\\"', '"')
            if as_dict:
                result[n] = v
            else:
                result.append((n, v))
    else:
        valid_data_len = 0
        for ws1, n, v, q1, vq, q2, ws2 in BP_EXPR_WS.findall(s):
            # print(",".join(["<"+x+">" for x in (ws1, n, v, q1, vq, q2, ws2)]))
            # check input
            if not n:
                raise BPParseError("Bad key: '{}'".format(n))
            valid_data_len += (
                sum(map(len, (ws1, n, v, q1, vq, q2, ws2))) + 1
            )  # 1 for '='
            # add to result
            if vq:
                v = vq.replace('\\"', '"')
            if as_dict:
                result[n] = v
            else:
                result.append((n, v))
        # check overall input
        junk_chars = len(s) - valid_data_len
        if junk_chars != 0:
            raise BPValidationError("{:d} junk chars in '{}'".format(junk_chars, s))
    return result


class ProcessInterface:
    """Process interface for a parser"""

    def process(self, line):
        """Subclasses must override this method to return a list
        of dictionaries or formatted log strings (with newlines).

        If there is an error with the format, they should
        raise a ValueError, KeyError, or this module's ParseError.
        If nothing is yet ready, return an empty list or tuple.
        To cause the caller to stop parsing this log, i.e. nothing will
        ever be ready, return None.
        """


class BaseParser(ProcessInterface, DoesLogging):
    """Base class for all other Parser classes in the parser modules.

    Uses iterator protocol to return one result at a time; where a
    result is a Best Practices log line (a string).
    Calls read(file) on the subclass to actually get and parse data.
    Each read() can return multiple events,
    or multiple read()s can return one event, transparently
    to the caller who only sees one log line per iteration.
    If 'unparsed_file' is not None, write all lines that returned an
    error, or the constant LINE_SKIPPED, to this file.

    Parameters:
      - add_hash {yes,no,no*}: To each output event, add a new field,
           '_hash', which is a probabilistically unique (MD5) hash of all
           the other fields in the event.
    """

    def __init__(
        self,
        input_file,
        fullname="unknown",
        unparsed_file=None,
        parse_date=True,
        add_hash="no",
        **kw
    ):
        """Initialize base parser.

        Parameters:

            input_file - File object (must support readline)

            fullname - For logging, the fully qualified name
                  for the logger (matches 'qualname' in the logging config).

            unparsed_file - File object to place records that caused
                  a parse exception

            parse_date - Whether to parse the ISO date to a number
                         or represent it as a string.

            **kw - Remaining keyword, value pairs are appended to each
                  line of the log. If the same keyword is in a
                  parsed result, the newer value takes precedence.
                  The exception to this is if the parser returns a string
                  instead of a dictionary, e.g. the 'bp' parser:
                  to avoid O(N*M) behavior where N is the number of
                  the keywords and M is the length of the output string,
                  duplicates are not checked.
        """
        if not input_file:
            raise ValueError("input file cannot be empty")
        DoesLogging.__init__(self, fullname)
        # common parameters
        self._add_hash = self.boolParam(add_hash)
        # rest of parameters
        self._infile = nlreadline.BufferedReadline(input_file)
        if hasattr(input_file, "fileno"):
            self._fake_file = False
            self._infile_rlist = (input_file.fileno(),)  # used in read_line
        else:
            # not a real file
            self._fake_file = True
            self._infile_rlist = ()
        try:
            self._offs = self._infile.tell()
        except OSError:
            self._offs = 0
        self._prev_len, self._saved_len = 0, 0
        self._saved = []
        self._name = fullname
        self._ufile = unparsed_file
        self._header_values = {}
        self._parser = NLSimpleParser(parse_date=parse_date)
        # Constant to add to each record
        self._const_nvp = {}
        # add GUID in env, if present
        guid = nlapi.getGuid(create=False)
        if guid:
            self._const_nvp["guid"] = guid
        # add user-provided values (can override guid)
        self._const_nvp.update(kw)
        # cache string-valued version, will be empty string if kw == {}
        self._const_nvp_str = " ".join(
            "{}={}".format(k, v) for k, v in self._const_nvp.items()
        )
        self.parse_date = parse_date

    def close(self):
        if self._infile:
            self._infile.close()

    def getFilename(self):
        if self._infile:
            return self._infile.name
        else:
            return ""

    def getOffset(self):
        """Return the offset of the last entirely parsed line.

        In the case of a single line that returned multiple items,
        all of which haven't yet been yet consumed, return the offset at
        the start of this line. This avoids dropping events at the
        expense of possible duplicates.

        It is best to call flush() first to avoid this issue entirely.

        """
        return self._offs

    def setOffset(self, offs):
        """Explicitly set offset.

        This is not normally necessary, as the next() function will
        advance self._offs every time all the resulting items from
        the associated input line have been returned.
        """
        self._infile.seek(offs)
        self._offs = offs

    def getParameters(self):
        """Subclasses should override this method to return
        a dictionary, with all basic types, representing any additional
        state that needs to be saved and restored.
        """
        return {}

    def setParameters(self, param):
        """Subclasses should override this method to update their
        state with the contents of the arg 'param', a dictionary.
        """

    def setHeaderValues(self, value_dict):
        """Set a dictionary of header keyword, value pairs."""
        self._header_values = value_dict

    def getHeaderValue(self, key):
        """Get value from group named 'key', or None."""
        return self._header_values.get(key, None)

    def __iter__(self):
        return self

    def next(self):
        """
        Return one saved or new item.

        Get new item(s) by reading and parsing the file.
        Return None if no result, so caller can
        count how many lines were processed and thus
        do fair sharing across multiple inputs
        """
        self.updateOffset()
        # get an item to return
        if self._saved:
            # multiple items were returned before, so just return one
            item = self._saved.pop()
            # if saved is now empty, then we have processed
            # all the items from the last readline, so
            # advance offset by its (saved) length
            if not self._saved:
                self._prev_len = self._saved_len
        else:
            line = self._read_line()
            # stop if line is empty
            if line == "":
                raise StopIteration
            item = line.strip()
            # main processing for the module
            try:
                result = self.process(item)
            except (ValueError, KeyError) as E:
                if self._ufile:
                    self._ufile.write(line)
                else:
                    self.log.warn("unparsed.event", value=line.strip(), msg=E)
                result = False
            # A skipped line means that it will never
            # be parsed, but there was no error.
            # Like an error, it is written to the unparsed-events
            # file (if that exists).
            if result == LINE_SKIPPED:
                if self._ufile:
                    self._ufile.write(line)
                result = False
            if not result:
                self._offs += len(line)
                if result is None:
                    raise StopIteration("EOF")
                item = None  # return this to caller
            else:
                item = result[0]
                if len(result) == 1:
                    # advance offset by this on next call
                    self._prev_len = len(line)
                else:
                    # don't advance offset until all results are returned
                    self._saved = list(result[1:])
                    self._saved.reverse()  # so we can pop()
                    self._saved_len = len(line)
        # return the item
        if item is None:
            return None
        else:
            return self._result(item)

    def _read_line(self):
        """Read one line."""
        if self._trace:
            self.log.trace("readline.start")
        if self._fake_file or select(self._infile_rlist, (), (), 0.1)[0]:
            line = self._infile.readline()
        else:
            line = ""
        if self._trace:
            self.log.trace("readline.end", n=len(line))
        return line

    def updateOffset(self):
        """Advance offset by length previously parsed input."""
        self._offs += self._prev_len
        self._prev_len = 0  # do not add this to offset again

    def _result(self, item):
        """Make item into a returnable result.
        Normalize 'level' and add constant attributes.
        Also run any postprocessing steps, such as adding a hash.
        """
        # Parse if a string
        if isinstance(item, str):
            item = self._parser.parseLine(item)
        # Normalize the 'level' value
        if "level" in item:
            level = item["level"]
            if hasattr(level, "upper"):
                lvlname = item["level"].upper()
                item["level"] = Level.getLevel(lvlname)
        # Add constant key, value pairs: do a copy and
        # reverse update so new values override old ones.
        if self._const_nvp:
            _tmp = self._const_nvp.copy()
            _tmp.update(item)
            item = _tmp
        # Do post-processing.
        if self._add_hash:
            item[HASH_FIELD] = hash_event(item)
        # Done
        return item

    def flush(self):
        """Return a list of all saved items, i.e., of all items
        that were parsed but not returned yet, and clear this list.
        """
        result = []
        # pending items, returned by parser
        for item in self._saved:
            result.append(self._result(item))
        self._saved = []
        self._offs += self._saved_len
        self._saved_len, self._prev_len = 0, 0
        return result

    def done(self):
        """Close internal state in parser.
        Useful if the parser is waiting for something, but hits EOF.
        """
        # items waiting in parser
        for item in self.finalize():
            self._saved.append(item)

    def finalize(self):
        """Any events to return at end of parse.
        Default is an empty list.
        """
        return ()

    def boolParam(self, s):
        """Convert a possibly-string-valued boolean parameter, from
        the configuration file, into a proper boolean value.
        """
        if isinstance(s, bool):
            return s
        if isinstance(s, str):
            sl = s.lower()
            if sl in ("yes", "on", "true", "1"):
                return True
            else:
                return False
        return bool(s)

    def __str__(self):
        return "{}({})".format(self._name, self._infile)


class NLBaseParser(BaseParser):
    def __init__(self, input_file=None, err_cb=None, **kw):
        """Create a NetLogger parser, that implements the BaseParser
        interface as well as its own API for parsing individual lines
        and streams with user-supplied callbacks.

        Arguments:
          input_file - File to be parsed
          err_cb - Optional callback on errors. Signature:
                     err_cb(line=<string with newline>, error=<Exception>,
                            linenum=<integer line number>)
                   If set to a function, the function is called.
                   If set to False, errors are completely ignored.
                   If set to True, errors are appended to err_list.
                   If None (the default) errors are propagated to the caller.
                           These errors are of type BPError.
        """
        self.err_cb = err_cb
        self.err_list = []
        if input_file is None:
            input_file = NullFile()
        BaseParser.__init__(self, input_file, fullname="NLParser", **kw)

    def parseLine(self, line):
        """Return a dictionary corresponding to the name,value pairs
        in the input 'line'.
        Raises ValueError if the format is incorrect.
        """

    def parseStream(self):
        """Parse input stream, calling parseLine() for each line.

        Return:
           generator function, yielding the result of parseLine() for
           each line in the input stream
        """
        for line_num, line in enumerate(self._infile):
            try:
                d = self.parseLine(line)
                yield d
            except ValueError as E:
                if self.err_cb is False:
                    pass
                else:
                    bpe = BPError(line_num, E)
                    if self.err_cb is True:
                        self.err_list.append(bpe)
                    elif self.err_cb is None:
                        raise bpe
                    else:
                        self.err_cb(line=line, error=bpe, linenum=line_num)


class NLSimpleParser(DoesLogging):
    """Simple, fast, not-too-flexible NL parser.

    Does *not* inherit from NLBaseParser.
    This is important to allow the BaseParser to itself parse
    netlogger input.
    """

    def __init__(self, verify=False, parse_date=True, **kw):
        DoesLogging.__init__(self)
        self.verify, self.parse_date = verify, parse_date

    def parseLine(self, line):
        """Parse a BP-formatted line.
        For lines which are completely whitespace, raise BPParseError
        """
        try:
            fields = _bp_extract(line, validate=self.verify)
        except BPError as err:
            raise BPParseError("BP parse error: " + str(err))
        # higher-level verification
        for key in TS_FIELD, EVENT_FIELD:
            if key not in fields:
                raise BPParseError("missing required key '{}'".format(key))
        # Pre-process date, if requested
        if self.parse_date:
            fields[TS_FIELD] = parse_ts(fields[TS_FIELD])
        # Done.
        return fields


class NLFastParser(NLSimpleParser, NLBaseParser):
    """NetLogger parser that does inherit from NLBaseParser
    Simpler, faster, less flexible NL parser.

    * Optionally does some error-checking, but can't tell why it is wrong.
    Note: error-checking takes 50% or so longer.
    * Uses regular expressions instead of pyparsing.
    * Observed speedups on order of 50x (YMMV).
    """

    def __init__(self, *args, **kw):
        # strip out kw for simple parser
        simple_parser_kw = {}
        for key, value in kw.items():
            if key in ("verify", "parse_date", "strip_quotes"):
                simple_parser_kw[key] = value
        for key in simple_parser_kw.keys():
            if key != "parse_date":  # shared
                del kw[key]
        NLSimpleParser.__init__(self, **simple_parser_kw)
        NLBaseParser.__init__(self, *args, **kw)

    # implementation of the BaseParser API
    def process(self, line):
        return (self.parseLine(line),)


if HAVE_PYPARSING:

    class NLPyParser(NLBaseParser):
        """pyparsing--based implementation of the NLBaseParser"""

        notSpace = CharsNotIn(" \n")
        eq = Literal("=").suppress()
        value = QuotedString('"', escChar=chr(92), unquoteResults=False) ^ OneOrMore(
            notSpace
        )
        ts = Group(Literal("ts") + eq + value)
        event = Group(Literal("event") + eq + value)
        name = ~oneOf("ts event") + Word(alphanums + "-_.")
        nv = ZeroOrMore(Group(name + eq + value))
        nvp = Each([ts, event, nv]) + White("\n").suppress() + StringEnd()

        def parseLine(self, line):
            try:
                rlist = self.nvp.parseString(line).asList()
            except ParseException as E:
                raise ValueError(E)
            result = {}
            for a in rlist:
                if self.parse_date and a[0] == "ts":
                    result[a[0]] = parse_ts(a[1])
                else:
                    result[a[0]] = a[1]
            return result

    # implementation of the BaseParser API
    def process(self, line):
        return (self.parseLine(line),)


else:

    class NLPyParser:
        BADNESS = """
Can't use the NLPyParser class because pyparsing is not installed.
You can use NLFastParser instead, run 'easy_install pyparsing', or
install from http://pyparsing.wikispaces.com/ .
"""

        def __init__(self, *args, **kw):
            raise NotImplementedError(self.BADNESS)


class NullFile:
    def __init__(self, *args):
        return

    def read(self, n):
        return ""

    def readline(self):
        return ""

    def seek(self, n, mode):
        pass

    def tell(self):
        return 0


def getTimezone(t=None):
    """Return current timezone as UTC offset,
    formatted as [+/-]HH:MM
    """
    hr, min, sign = nldate.getLocaltimeOffsetParts(t)
    return "%s%02d:%02d" % (sign, hr, min)
